/**********************************************************************************************************************************/
/* Project      : TAPER                                                                                                           */
/* Module       : TPCONFIG.C                                                                                                      */
/* Description  : Configuration handler                                                                                           */
/* Version type : Program module                                                                                                  */
/* Last changed : 09-12-2001  11:00                                                                                               */
/* Update count : 42                                                                                                              */
/* OS type      : PC (DOS)                                                                                                        */
/* Compiler     : Watcom C32 Optimizing Compiler Version 10.x                                                                     */
/* Linker       : Watcom C Linker                                                                                                 */
/*                                                                                                                                */
/*                Copyleft (C) 1997-2001 ThunderWare Research Center, written by Martijn van der Heide.                           */
/*                                                                                                                                */
/*                This program is free software; you can redistribute it and/or                                                   */
/*                modify it under the terms of the GNU General Public License                                                     */
/*                as published by the Free Software Foundation; either version 2                                                  */
/*                of the License, or (at your option) any later version.                                                          */
/*                                                                                                                                */
/*                This program is distributed in the hope that it will be useful,                                                 */
/*                but WITHOUT ANY WARRANTY; without even the implied warranty of                                                  */
/*                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                                   */
/*                GNU General Public License for more details.                                                                    */
/*                                                                                                                                */
/*                You should have received a copy of the GNU General Public License                                               */
/*                along with this program; if not, write to the Free Software                                                     */
/*                Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                     */
/*                                                                                                                                */
/**********************************************************************************************************************************/

#define __TPCONFIG_MAIN__

#include "wbench32/wbench32.h"
#include <dos.h>
#include <fcntl.h>
#include <io.h>
#include "tpmain.h"
#include "tpconfig.h"

#pragma pack                ()

#include "speclogo.inc"

#define REMAP               "RemapBrightBlack"
#define DOFLASH             "EmulateFlash"
#define SLOWFLASH           "SlowFlash"
#define CONFQUIT            "ConfirmQuit"
#define CONFDEL             "ConfirmDelete"
#define CONFOVERW           "ConfirmOverwrite"
#define SYNCSLID            "FollowSlider"
#define EDITINS             "EditInsert"
#define BACKUPTAPE          "BackupTape"
#define NOSOUNDS            "DisableSounds"
#define TRACESOUND          "TraceSoundBlocks"
#define DEFFILTER           "DefaultVOCFilter"
#define EDITOR              "Editor"
#define ZXEDIT              "ZX-EditPath"
#define GIFVIEWER           "GIFViewerPath"
#define GIFOPTIONS          "GIFViewerOptions"
#define JPGVIEWER           "JPGViewerPath"
#define JPGOPTIONS          "JPGViewerOptions"
#define PRNPORT             "LPTPort"
#define COMPORT             "COMPort"
#define PLAYSOUNDB          "PlayDeviceSoundBlaster"
#define PLAYCRIT            "PlayWindowsDisabled"
#define PLAYPIRQ            "ParPlayIRQsDisabled"
#define SBNOHIDMA           "SBNoHighSpeedDMA"
#define SAMPVALUES          "MixerInputValues"
#define BLANKTO             "BlankerTimeOut"
#define DRAGOUTLN           "DragOutlineOnly"
#define RESZOUTLN           "ResizeOutlineOnly"
#define GREYGHOST           "GreyOutGhosting"
#define PRELOADRES          "PreLoadResources"
#define DISABLERIL          "DirectVideoControl"
#define SWAPMODE80          "SwapInTextMode"
#define QUITMODE80          "QuitInTextMode"
#define NEXTDISKID          "NextFloppyDiskId"
#define GMOUSE              "UseGeniusMouseHiding"
#define SWAPBUTT            "SwapMouseButtons"
#define EMUL3BUTT           "Emulate3rdMouseButton"
#define NOLOCHIDE           "NoLocalMouseHiding"
#define FORCEMOUSE          "ForceMouseRelease"
#define DCLICKDELAY         "DoubleClickInterval"
#define CPUSHDELAY          "InitialContPushDelay"
#define MOUSESPEED          "MouseSpeedSetting"
#define SNAPMOUSE           "SnapMouseDefault"
#define DEFDECROM           "DefaultDecodeROMTiming"
#define TZXSTRPAUSE         "TZXStripPauseLengths"
#define TRACETIME           "ShowRemainingPlayTime"
#define CUSTOMPAL           "UseCustomPalette"
#define CUSTOMCOL00         "CustomColor00"
#define CUSTOMCOL01         "CustomColor01"
#define CUSTOMCOL02         "CustomColor02"
#define CUSTOMCOL03         "CustomColor03"
#define CUSTOMCOL04         "CustomColor04"
#define CUSTOMCOL05         "CustomColor05"
#define CUSTOMCOL06         "CustomColor06"
#define CUSTOMCOL07         "CustomColor07"
#define CUSTOMCOL08         "CustomColor08"
#define CUSTOMCOL09         "CustomColor09"
#define CUSTOMCOL10         "CustomColor10"
#define CUSTOMCOL11         "CustomColor11"
#define CUSTOMCOL12         "CustomColor12"
#define CUSTOMCOL13         "CustomColor13"
#define CUSTOMCOL14         "CustomColor14"
#define CUSTOMCOL15         "CustomColor15"

/**********************************************************************************************************************************/
/* Define the static variables                                                                                                    */
/**********************************************************************************************************************************/

static struct ConfOp_s                                                                  /* The commands available in a CONFIGFILE */
{
  char *OptionName;
  enum
  {
    OPT_BOOL,
    OPT_WORD,
    OPT_HWORD,
    OPT_BYTE,
    OPT_DWORD,
    OPT_STRING,
    OPT_STR15,
    OPT_PATH,
    OPT_IGNORE
  }     OptionType;
  void *DataPtr;
};

static struct ConfOp_s     ConfOp[] = { { DOFLASH,     OPT_BOOL,   &TPConfig.EmulateFlash          },
                                        { SLOWFLASH,   OPT_BOOL,   &TPConfig.SlowFlash             },
                                        { REMAP,       OPT_BOOL,   &TPConfig.RemapBrightBlack      },
                                        { CONFQUIT,    OPT_BOOL,   &TPConfig.ConfirmQuit           },
                                        { CONFDEL,     OPT_BOOL,   &TPConfig.ConfirmDelete         },
                                        { CONFOVERW,   OPT_BOOL,   &TPConfig.ConfirmOverwrite      },
                                        { SYNCSLID,    OPT_BOOL,   &SystemDefaults.SyncSlider      },
                                        { EDITINS,     OPT_BOOL,   &SystemDefaults.EditInsert      },
                                        { BACKUPTAPE,  OPT_BOOL,   &TPConfig.AlwaysBackupTape      },
                                        { NOSOUNDS,    OPT_BOOL,   &SystemDefaults.Quiet           },
                                        { TRACESOUND,  OPT_BOOL,   &TPConfig.TraceSoundBlocks      },
                                        { TRACETIME,   OPT_BOOL,   &TPConfig.ShowRemainingPlayTime },
                                        { PLAYPIRQ,    OPT_BOOL,   &TPConfig.ParPlayIRQsDisabled   },
                                        { SBNOHIDMA,   OPT_BOOL,   &TPConfig.SBNoHighSpeedDMA      },
                                        { SAMPVALUES,  OPT_STR15,   TPConfig.SampleValues          },
                                        { DEFFILTER,   OPT_BOOL,   &TPConfig.DefUseFilter          },
                                        { EDITOR,      OPT_PATH,    TPConfig.EditPath              },
                                        { ZXEDIT,      OPT_PATH,    TPConfig.ZXEditPath            },
                                        { GIFVIEWER,   OPT_PATH,    TPConfig.GIFViewerPath         },
                                        { JPGVIEWER,   OPT_PATH,    TPConfig.JPGViewerPath         },
                                        { GIFOPTIONS,  OPT_STRING,  TPConfig.GIFViewerOptions      },
                                        { JPGOPTIONS,  OPT_STRING,  TPConfig.JPGViewerOptions      },
                                        { PRNPORT,     OPT_BYTE,   &TPConfig.LPTPort               },
                                        { COMPORT,     OPT_BYTE,   &TPConfig.COMPort               },
                                        { BLANKTO,     OPT_WORD,   &TPConfig.BlankerTimeOut        },
                                        { DRAGOUTLN,   OPT_BOOL,   &SystemDefaults.DragOutline     },
                                        { RESZOUTLN,   OPT_BOOL,   &SystemDefaults.ResizeOutline   },
                                        { GREYGHOST,   OPT_BOOL,   &SystemDefaults.GreyOutGhosting },
                                        { PRELOADRES,  OPT_BOOL,   &TPConfig.PreLoadResources      },
                                        { DISABLERIL,  OPT_BOOL,   &TPConfig.DirectVideoControl    },
                                        { SWAPMODE80,  OPT_BOOL,   &SystemDefaults.ForceSwapMode80 },
                                        { QUITMODE80,  OPT_BOOL,   &SystemDefaults.ForceQuitMode80 },
                                        { DEFDECROM,   OPT_BOOL,   &TPConfig.DefDecodeROM          },
                                        { TZXSTRPAUSE, OPT_BOOL,   &TPConfig.TZXStripPauseLengths  },
                                        { PLAYSOUNDB,  OPT_BOOL,   &TPConfig.PlaySoundBlaster      },
                                        { PLAYCRIT,    OPT_BOOL,   &TPConfig.PlayCriticalSection   },
                                        { NEXTDISKID,  OPT_WORD,   &NextFloppyId                   },
                                        { GMOUSE,      OPT_BOOL,   &Mouse.UseGeniusScreenUpdate    },
                                        { NOLOCHIDE,   OPT_BOOL,   &Mouse.DisableLocalHiding       },
                                        { FORCEMOUSE,  OPT_BOOL,   &Mouse.ForceMouseRelease        },
                                        { SWAPBUTT,    OPT_BOOL,   &Mouse.SwapLeftAndRightButton   },
                                        { EMUL3BUTT,   OPT_BOOL,   &Mouse.EmulateMiddleMouseButton },
                                        { SNAPMOUSE,   OPT_BOOL,   &Mouse.SnapToDefaultButton      },
                                        { DCLICKDELAY, OPT_WORD,   &Mouse.DoubleClickInterval      },
                                        { CPUSHDELAY,  OPT_WORD,   &Mouse.CPushDelay               },
                                        { MOUSESPEED,  OPT_STRING,  TPConfig.MouseSpeedString      },
                                        { CUSTOMPAL,   OPT_BOOL,   &TPConfig.UseCustomPalette      },
                                        { CUSTOMCOL00, OPT_DWORD,  &TPConfig.CustomColor00         },
                                        { CUSTOMCOL01, OPT_DWORD,  &TPConfig.CustomColor01         },
                                        { CUSTOMCOL02, OPT_DWORD,  &TPConfig.CustomColor02         },
                                        { CUSTOMCOL03, OPT_DWORD,  &TPConfig.CustomColor03         },
                                        { CUSTOMCOL04, OPT_DWORD,  &TPConfig.CustomColor04         },
                                        { CUSTOMCOL05, OPT_DWORD,  &TPConfig.CustomColor05         },
                                        { CUSTOMCOL06, OPT_DWORD,  &TPConfig.CustomColor06         },
                                        { CUSTOMCOL07, OPT_DWORD,  &TPConfig.CustomColor07         },
                                        { CUSTOMCOL08, OPT_DWORD,  &TPConfig.CustomColor08         },
                                        { CUSTOMCOL09, OPT_DWORD,  &TPConfig.CustomColor09         },
                                        { CUSTOMCOL10, OPT_DWORD,  &TPConfig.CustomColor10         },
                                        { CUSTOMCOL11, OPT_DWORD,  &TPConfig.CustomColor11         },
                                        { CUSTOMCOL12, OPT_DWORD,  &TPConfig.CustomColor12         },
                                        { CUSTOMCOL13, OPT_DWORD,  &TPConfig.CustomColor13         },
                                        { CUSTOMCOL14, OPT_DWORD,  &TPConfig.CustomColor14         },
                                        { CUSTOMCOL15, OPT_DWORD,  &TPConfig.CustomColor15         },
                                        { "LPTAddress1", OPT_IGNORE, NULL                          },
                                        { "LPTAddress2", OPT_IGNORE, NULL                          },
                                        { "LPTAddress3", OPT_IGNORE, NULL                          },
                                        { "COMAddress1", OPT_IGNORE, NULL                          },
                                        { "COMAddress2", OPT_IGNORE, NULL                          },
                                        { "COMAddress3", OPT_IGNORE, NULL                          },
                                        { "COMAddress4", OPT_IGNORE, NULL                          } };

#define NumOps              (sizeof (ConfOp) / sizeof (struct ConfOp_s))

/**********************************************************************************************************************************/
/* Define the static routines                                                                                                     */
/**********************************************************************************************************************************/

static int  _NextString     (FILE *Fp, char *String, bool FindEQual, bool SkipEOL);
static void _GetNewFileName (char *Operation, char *OldName);

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE INTERNAL GRAPHICS INTERFACE LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

static int _NextString (FILE *Fp, char *String, bool FindEqual, bool SkipEOL)

/**********************************************************************************************************************************/
/* Pre   : 'Fp' points to the open CONFIGFILE, 'String' holds the address of the string to be read, 'FindEqual' is TRUE if the    */
/*         first character to be found must be a '=' sign, 'SkipEOL' is TRUE if '\n' characters may be skipped.                   */
/* Post  : Any string is read. If it started with a '#' (COMMENT character), then the rest of the line has been ignored. The      */
/*         function does not return before a string was read WITHOUT this character or EOF has been reached. The result of the    */
/*         function is the length of the read string, or 0 if EOF was encountered.                                                */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  char          Ch;
  char register Cnt;
  bool          Ready = FALSE;
  bool          SkipSpaces;

  String[0] = '\0';
  while (!Ready)                                                                            /* Read until a valid string is found */
  {
    SkipSpaces = TRUE;
    while (SkipSpaces)
    {
      fscanf (Fp, "%c", &Ch);                                                                 /* Read until no white-spaces found */
      if (feof (Fp))
      {
        String[0] = '\0';                                                                                         /* Or until EOF */
        return (0);
      }
      SkipSpaces = isspace (Ch);
      if (Ch == '\n' && !SkipEOL)                                               /* EOL character found which may not be skipped ? */
      {
        String[0] = '\0';                                                                        /* Then return the EOL character */
        return (0);
      }                                                                               /* (Otherwise it is a whitespace character) */
    }
    if (Ch == '#')                                                                  /* First character is the COMMENT character ? */
    {
      do
      {
        fscanf (Fp, "%c", &Ch);                                                                   /* Ignore until end of the line */
        if (feof (Fp))
        {
          String[0] = '\0';                                                                                       /* Or until EOF */
          return (0);
        }
      }
      while (Ch != '\n');
      if (!SkipEOL)                                                                            /* Do not skip the EOL character ? */
      {
        String[0] = '\0';                                                                        /* Then return the EOL character */
        return (0);
      }
    }
    else
      Ready = TRUE;
  }
  Cnt = 0;
  Ready = FALSE;
  if (FindEqual)
    return (Ch == '=');
  while (!Ready)
  {
    while (!isspace (Ch) && Ch != '"')                                                             /* Trap quoted argument(part)s */
    {
      String[Cnt ++] = Ch;
      fscanf (Fp, "%c", &Ch);                                                                     /* Read until first white-space */
      if (feof (Fp))
      {
        String[Cnt] = '\0';                                                                                       /* Or until EOF */
        return (Cnt);
      }
    }
    if (Ch == '"')                                                                                          /* Handle quoted part */
    {
      do
      {
        fscanf (Fp, "%c", &Ch);
        if (feof (Fp))
          FatalError (11, "Config: Unexpected end of configuration file");
        if (Ch == '\n')
          FatalError (11, "Config: Unexpected end of line in configuration file");
        String[Cnt ++] = Ch;
      }
      while (Ch != '"');
      Cnt --;
      fscanf (Fp, "%c", &Ch);                                                                                 /* Read first after */
      if (feof (Fp))
      {
        String[Cnt] = '\0';                                                                                       /* Or until EOF */
        return (Cnt);
      }
    }
    else
      Ready = TRUE;
  }
  String[Cnt] = '\0';
  return (Cnt);
}

static void _GetNewFileName (char *Operation, char *OldName)

/**********************************************************************************************************************************/
/* Pre   : `Operation' is a string for the file requester header, `OldName' is the current path name.                             */
/* Post  : A new filename has been asked with a FileRequester, which is returned in `OldName'.                                    */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  char           Path[MAXDIR + 1];
  char           File[13];
  char           Mask[13] = "*.*";
  short register Cnt;

  if (OldName[0] != '\0')                                                                              /* A path has been given ? */
  {
    Cnt = strlen (OldName);
    while (-- Cnt >= 0 && OldName[Cnt] != '\\' && OldName[Cnt] != ':')
      ;
    if (Cnt >= 0)                                                                                    /* Is there a leading path ? */
    {
      strncpy (Path, OldName, Cnt);
      Path[Cnt] = '\0';
    }
    else
      Path[0] = '\0';
    strcpy (File, OldName + Cnt + 1);
  }
  else
    Path[0] = File[0] = '\0';
  if (FileRequester (180, 50, 48, 29, 20, Path, File, Mask, Operation, TRUE, FALSE))
    MakeFullPath (OldName, Path, File);
  if (MaxFloppyIdChanged)
    WriteConfig ();
}

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE EXPORTED LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

void ReadConfig (void)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : If the configuration file is found, then it has been read out. Otherwise, one is created.                              */
/*         If one (or all) are not found, then they are initialized with the defaults.                                            */
/*         If a character '#' is encountered, then the rest of this line is ignored (comment).                                    */
/* Import: _NextString.                                                                                                           */
/**********************************************************************************************************************************/

{
  FILE         *Fp;
  char          Item[256];
  char         *String;
  char          S[256];
  bool          Handled;
  byte register CntOp;
  byte register N;
  int           Len;
  bool          PathError;
  bool          Finished;
  short         Val;
  short         ValNo;
  short         HorizontalMickeys;
  short         VerticalMickeys;
  short         DoubleSpeedThreshold;
  char         *Ch;
  
  if (IsFloppy (ProgramDrive[0]))
    RequestFloppy (ProgramDrive[0], ResourceFloppyId);
  _InvalidPort = FALSE;
  TPConfig.RemapBrightBlack = TRUE;
  TPConfig.EmulateFlash = TRUE;
  TPConfig.SlowFlash = FALSE;
  TPConfig.TraceSoundBlocks = TRUE;
  TPConfig.ShowRemainingPlayTime = TRUE;
  TPConfig.DefUseFilter = TRUE;
  TPConfig.ConfirmQuit = TRUE;
  TPConfig.ConfirmDelete = TRUE;
  TPConfig.ConfirmOverwrite = TRUE;
  SystemDefaults.EditInsert = FALSE;
  TPConfig.AlwaysBackupTape = TRUE;
  SystemDefaults.Quiet = FALSE;
  TPConfig.EditPath[0] = TPConfig.ZXEditPath[0] = TPConfig.GIFViewerPath[0] = TPConfig.JPGViewerPath[0] =
  TPConfig.GIFViewerOptions[0] = TPConfig.JPGViewerOptions[0] = '\0';
  TPConfig.DefDecodeROM = FALSE;
  TPConfig.TZXStripPauseLengths = TRUE;
  TPConfig.LPTPort = 1;
  TPConfig.COMPort = 1;
  TPConfig.COMAddr[0] = *((word *)0x00000400);
  TPConfig.COMAddr[1] = *((word *)0x00000402);
  TPConfig.COMAddr[2] = *((word *)0x00000404);
  TPConfig.COMAddr[3] = *((word *)0x00000406);
  TPConfig.LPTAddr[0] = *((word *)0x00000408);
  TPConfig.LPTAddr[1] = *((word *)0x0000040A);
  TPConfig.LPTAddr[2] = *((word *)0x0000040C);
  TPConfig.PlaySoundBlaster = TRUE;
  TPConfig.PlayCriticalSection = FALSE;
  TPConfig.ParPlayIRQsDisabled = FALSE;
  TPConfig.SBNoHighSpeedDMA = FALSE;
  strcpy (TPConfig.SampleValues, "0,0,0,0");
  TPConfig.BlankerOn = TRUE;
  TPConfig.BlankerTimeOut = 600;
  TPConfig.DirectVideoControl = TRUE;
  TPConfig.PreLoadResources = FALSE;
  NextFloppyId = 1;
  TPConfig.UseCustomPalette = FALSE;
  TPConfig.CustomColor00 = 0x00000000;   /* VGA  0: DBLACK   */
  TPConfig.CustomColor01 = 0x002F0000;   /* VGA  1: DBLUE    */
  TPConfig.CustomColor02 = 0x0000002F;   /* VGA  4: DRED     */
  TPConfig.CustomColor03 = 0x002F002F;   /* VGA  5: DMAGENTA */
  TPConfig.CustomColor04 = 0x00002F00;   /* VGA  2: DGREEN   */
  TPConfig.CustomColor05 = 0x002F2F00;   /* VGA  3: DCYAN    */
  TPConfig.CustomColor06 = 0x00002F2F;   /* VGA  6: DYELLOW  */
  TPConfig.CustomColor07 = 0x002A2A2A;   /* VGA  7: DWHITE   */
  TPConfig.CustomColor08 = 0x001A1A1A;   /* VGA  8: LBLACK   */
  TPConfig.CustomColor09 = 0x003F0000;   /* VGA  9: LBLUE    */
  TPConfig.CustomColor10 = 0x0000003F;   /* VGA 12: LRED     */
  TPConfig.CustomColor11 = 0x003F003F;   /* VGA 13: LMAGENTA */
  TPConfig.CustomColor12 = 0x00003F00;   /* VGA 10: LGREEN   */
  TPConfig.CustomColor13 = 0x003F3F00;   /* VGA 11: LCYAN    */
  TPConfig.CustomColor14 = 0x00003F3F;   /* VGA 14: LYELLOW  */
  TPConfig.CustomColor15 = 0x003F3F3F;   /* VGA 15: LWHITE   */
  sprintf (TPConfig.MouseSpeedString, "%u,%u,%u",Mouse.HorizontalMickeys, Mouse.VerticalMickeys, Mouse.DoubleSpeedThreshold);
  if ((Fp = fopen (ConfigFile, "r")) == NULL)                                                      /* Configuration file exists ? */
  {
    WriteConfig ();                                                                                             /* No, create one */
    return;
  }
  Len = _NextString (Fp, Item, FALSE, TRUE);                                        /* Read-ahead first string: must be a keyword */
  while (!feof (Fp))
  {
    if (Item[Len - 1] == '=')
      Item[Len - 1] = '\0';
    else
      if (!_NextString (Fp, S, TRUE, FALSE))
        FatalError (11, "Config: Missing \'=\' sign in option %s", Item);
    Handled = FALSE;
    for (CntOp = 0 ; !Handled && CntOp < NumOps ; CntOp ++)
      if (!stricmp (Item, ConfOp[CntOp].OptionName))
      {
        _NextString (Fp, Item, FALSE, FALSE);                                                                 /* Read the operand */
        if (Item[0] == '\0')                                                                                /* An empty operand ? */
        {
          if (ConfOp[CntOp].OptionType == OPT_BYTE || ConfOp[CntOp].OptionType == OPT_WORD || ConfOp[CntOp].OptionType == OPT_BOOL)
             FatalError (11, "Config: Missing operand after switch %s", ConfOp[CntOp].OptionName);
        }
        else
          switch (ConfOp[CntOp].OptionType)                                                                   /* Parse the option */
          {
            case OPT_BYTE    :
            case OPT_WORD    : N = 0;
                               if (!isdigit (Item[N]))
                                 if (Item[N ++] != '-')
                                   FatalError (11, "Config: Invalid number %s after switch %s", Item, ConfOp[CntOp].OptionName);
                               for ( ; N < strlen (Item) ; N ++)
                                 if (!isdigit (Item[N]))
                                   FatalError (11, "Config: Invalid number %s after switch %s", Item, ConfOp[CntOp].OptionName);
                               if (ConfOp[CntOp].OptionType == OPT_WORD)
                                 *((word *)ConfOp[CntOp].DataPtr) = atoi (Item);
                               else
                                 *((byte *)ConfOp[CntOp].DataPtr) = atoi (Item);
                               break;
            case OPT_HWORD   : if (Item[0] != '0' || toupper (Item[1]) != 'X')
                                 FatalError (11, "Config: Invalid hex number %s after switch %s", Item, ConfOp[CntOp].OptionName);
                               *((word *)ConfOp[CntOp].DataPtr) = 0;
                               for (N = 2 ; N < strlen (Item) ; N ++)
                                 if (!isxdigit (Item[N]))
                                   FatalError (11, "Config: Invalid hex number %s after switch %s", Item, ConfOp[CntOp].OptionName);
                                 else
                                 {
                                   *((word *)ConfOp[CntOp].DataPtr) *= 16;
                                   if (isdigit (Item[N]))
                                     *((word *)ConfOp[CntOp].DataPtr) += (Item[N] - '0');
                                   else
                                     *((word *)ConfOp[CntOp].DataPtr) += (toupper (Item[N]) - 'A' + 10);
                                 }
                               break;
            case OPT_DWORD   : if (Item[0] != '0' || toupper (Item[1]) != 'X')
                                 FatalError (11, "Config: Invalid hex number %s after switch %s", Item, ConfOp[CntOp].OptionName);
                               *((dword *)ConfOp[CntOp].DataPtr) = 0;
                               for (N = 2 ; N < strlen (Item) ; N ++)
                                 if (!isxdigit (Item[N]))
                                   FatalError (11, "Config: Invalid hex number %s after switch %s", Item, ConfOp[CntOp].OptionName);
                                 else
                                 {
                                   *((dword *)ConfOp[CntOp].DataPtr) *= 16;
                                   if (isdigit (Item[N]))
                                     *((dword *)ConfOp[CntOp].DataPtr) += (Item[N] - '0');
                                   else
                                     *((dword *)ConfOp[CntOp].DataPtr) += (toupper (Item[N]) - 'A' + 10);
                                 }
                               break;
            case OPT_BOOL    : if (!stricmp (Item, "On") || !stricmp (Item, "TRUE") || !strcmp (Item, "1") || !stricmp (Item,"Yes"))
                                 *((bool *)ConfOp[CntOp].DataPtr) = TRUE;
                               else if (!stricmp (Item, "Off") || !stricmp (Item, "FALSE") || !strcmp (Item, "0") ||
                                        !stricmp (Item, "No"))
                                 *((bool *)ConfOp[CntOp].DataPtr) = FALSE;
                               else
                                 FatalError (11, "Config: Unknown boolean type %s", Item);
                               break;
            case OPT_STR15   : if (strlen (Item) > 15)
                                 FatalError (11, "Config: %s: \"%s\" too long", ConfOp[CntOp].OptionName, Item);
                               strcpy ((char *)ConfOp[CntOp].DataPtr, Item);
                               break;
            case OPT_STRING  : strcpy ((char *)ConfOp[CntOp].DataPtr, Item);
                               break;
            case OPT_PATH    : String = TestName (Item, FALSE, &PathError);
                               if (PathError)
                                 FatalError (11, "Config: %s", String);
                               strcpy ((char *)ConfOp[CntOp].DataPtr, String);
                               strupr ((char *)ConfOp[CntOp].DataPtr);
                               break;
          }
        Handled = TRUE;
      }
    if (!Handled)                                                                         /* Read item is not one of the keywords */
      FatalError (11, "Config: Unknown keyword %s in configuration file", Item);
    _NextString (Fp, Item, FALSE, TRUE);                                                                      /* Read next option */
    if (TPConfig.BlankerTimeOut == (word)-1)
    {
      TPConfig.BlankerOn = FALSE;
      TPConfig.BlankerTimeOut = 600;
    }
  }
  fclose (Fp);
  if (TPConfig.LPTPort < 1 || TPConfig.LPTPort > 3)
  { _InvalidPort = TRUE; TPConfig.LPTPort = 1; }
  else if (TPConfig.LPTPort != 1 && TPConfig.LPTAddr[TPConfig.LPTPort - 1] == 0x0000)                       /* Non-existing port! */
  { _InvalidPort = TRUE; TPConfig.LPTPort = 1; }
  if (TPConfig.COMPort < 1 || TPConfig.COMPort > 4)
  { _InvalidPort = TRUE; TPConfig.COMPort = 1; }
  else if (TPConfig.COMPort != 1 && TPConfig.COMAddr[TPConfig.COMPort - 1] == 0x0000)
  { _InvalidPort = TRUE; TPConfig.COMPort = 1; }
  TapeWindow[TWLEFT].BackupTape = TapeWindow[TWRIGHT].BackupTape = TPConfig.AlwaysBackupTape;
  if (Mouse.UseGeniusScreenUpdate && Mouse.DisableLocalHiding)
    Mouse.UseGeniusScreenUpdate = FALSE;
  SoundBlasterInfo.CanDoHighSpeedDMA = !TPConfig.SBNoHighSpeedDMA;
  Ch = TPConfig.MouseSpeedString;
  Val = 0;
  ValNo = 0;
  while (!Finished)                                           /* First parse the configuration setting to set the register values */
  {
    if (isdigit (*Ch))
      Val = Val * 10 + *Ch - '0';
    else if (*Ch == ',' || *Ch == '\0')
    {
      if (Val != 0)
        switch (ValNo)
        {
          case 0 : HorizontalMickeys = Val; break;
          case 1 : VerticalMickeys = Val; break;
          case 2 : DoubleSpeedThreshold = Val; break;
        }
      Val = 0;
      ValNo ++;
      if (*Ch == '\0')
        Finished = TRUE;
    }
    Ch ++;
  }
  SetMouseSpeed (HorizontalMickeys, VerticalMickeys, DoubleSpeedThreshold);
}

void WriteConfig (void)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The configuration has been written to disk.                                                                            */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  FILE          *Fp;

  if (IsFloppy (ProgramDrive[0]))
    RequestFloppy (ProgramDrive[0], ResourceFloppyId);
  if ((Fp = fopen (ConfigFile, "w")) == NULL)
    FatalError (11, "Cannot write configuration file %s", ConfigFile);
  fprintf (Fp, "# %s%s - Configuration of %s\n", ProgramName, CONFIGEXT, ProgramName);
  fprintf (Fp, "# Copyleft (C) 1997-2001 ThunderWare Research Center\n#\n");
  fprintf (Fp, "# These items are updated from the configuration window\n\n");
  fprintf (Fp, "%-22s = %s\n", REMAP,          TPConfig.RemapBrightBlack ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", DOFLASH,        TPConfig.EmulateFlash ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", SLOWFLASH,      TPConfig.SlowFlash ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", CONFQUIT,       TPConfig.ConfirmQuit ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", CONFDEL,        TPConfig.ConfirmDelete ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", CONFOVERW,      TPConfig.ConfirmOverwrite ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", SYNCSLID,       SystemDefaults.SyncSlider ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", EDITINS,        SystemDefaults.EditInsert ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", BACKUPTAPE,     TPConfig.AlwaysBackupTape ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", NOSOUNDS,       SystemDefaults.Quiet ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", TRACESOUND,     TPConfig.TraceSoundBlocks ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", TRACETIME,      TPConfig.ShowRemainingPlayTime ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", DEFFILTER,      TPConfig.DefUseFilter ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", EDITOR,         TPConfig.EditPath);
  fprintf (Fp, "%-22s = %d\n", PRNPORT,        TPConfig.LPTPort);
  fprintf (Fp, "%-22s = %d\n", COMPORT,        TPConfig.COMPort);
  fprintf (Fp, "%-22s = %s\n", PLAYSOUNDB,     TPConfig.PlaySoundBlaster ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", PLAYCRIT,       TPConfig.PlayCriticalSection ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", PLAYPIRQ,       TPConfig.ParPlayIRQsDisabled ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", SAMPVALUES,     TPConfig.SampleValues);
  fprintf (Fp, "%-22s = %d\n", BLANKTO,        TPConfig.BlankerOn ? TPConfig.BlankerTimeOut : -1);
  fprintf (Fp, "\n# These items must be changed by hand");
  fprintf (Fp, "\n# General environment\n\n");
  fprintf (Fp, "%-22s = %s\n", ZXEDIT,         TPConfig.ZXEditPath);
  fprintf (Fp, "%-22s = %s\n", GIFVIEWER,      TPConfig.GIFViewerPath);
  fprintf (Fp, "%-22s = \"%s\"\n", GIFOPTIONS, TPConfig.GIFViewerOptions);
  fprintf (Fp, "%-22s = %s\n", JPGVIEWER,      TPConfig.JPGViewerPath);
  fprintf (Fp, "%-22s = \"%s\"\n", JPGOPTIONS, TPConfig.JPGViewerOptions);
  fprintf (Fp, "%-22s = %s\n", DRAGOUTLN,      SystemDefaults.DragOutline ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", RESZOUTLN,      SystemDefaults.ResizeOutline ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", GREYGHOST,      SystemDefaults.GreyOutGhosting ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", PRELOADRES,     TPConfig.PreLoadResources ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", DISABLERIL,     TPConfig.DirectVideoControl ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", SWAPMODE80,     SystemDefaults.ForceSwapMode80 ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", QUITMODE80,     SystemDefaults.ForceQuitMode80 ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", DEFDECROM,      TPConfig.DefDecodeROM ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", SBNOHIDMA,      TPConfig.SBNoHighSpeedDMA ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", TZXSTRPAUSE,    TPConfig.TZXStripPauseLengths ? "On" : "Off");
  fprintf (Fp, "%-22s = %d   # Don't change this one by hand!\n", NEXTDISKID,    NextFloppyId);
  fprintf (Fp, "\n# Mouse options\n\n");
  fprintf (Fp, "%-22s = %s\n", GMOUSE,         Mouse.UseGeniusScreenUpdate ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", NOLOCHIDE,      Mouse.DisableLocalHiding ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", FORCEMOUSE,     Mouse.ForceMouseRelease ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", SWAPBUTT,       Mouse.SwapLeftAndRightButton ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", EMUL3BUTT,      Mouse.EmulateMiddleMouseButton ? "On" : "Off");
  fprintf (Fp, "%-22s = %s\n", SNAPMOUSE,      Mouse.SnapToDefaultButton ? "On" : "Off");
  fprintf (Fp, "%-22s = %ld\n", DCLICKDELAY,   Mouse.DoubleClickInterval);
  fprintf (Fp, "%-22s = %ld\n", CPUSHDELAY,    Mouse.CPushDelay);
  fprintf (Fp, "%-22s = %s\n", MOUSESPEED,     TPConfig.MouseSpeedString);
  fprintf (Fp, "\n# Change these only if you know what you're doing!\n\n");
  fprintf (Fp, "%-22s = %s\n", CUSTOMPAL,     TPConfig.UseCustomPalette ? "On" : "Off");
  fprintf (Fp, "%-22s = 0x%06lX\n", CUSTOMCOL00, TPConfig.CustomColor00);
  fprintf (Fp, "%-22s = 0x%06lX\n", CUSTOMCOL01, TPConfig.CustomColor01);
  fprintf (Fp, "%-22s = 0x%06lX\n", CUSTOMCOL02, TPConfig.CustomColor02);
  fprintf (Fp, "%-22s = 0x%06lX\n", CUSTOMCOL03, TPConfig.CustomColor03);
  fprintf (Fp, "%-22s = 0x%06lX\n", CUSTOMCOL04, TPConfig.CustomColor04);
  fprintf (Fp, "%-22s = 0x%06lX\n", CUSTOMCOL05, TPConfig.CustomColor05);
  fprintf (Fp, "%-22s = 0x%06lX\n", CUSTOMCOL06, TPConfig.CustomColor06);
  fprintf (Fp, "%-22s = 0x%06lX\n", CUSTOMCOL07, TPConfig.CustomColor07);
  fprintf (Fp, "%-22s = 0x%06lX\n", CUSTOMCOL08, TPConfig.CustomColor08);
  fprintf (Fp, "%-22s = 0x%06lX\n", CUSTOMCOL09, TPConfig.CustomColor09);
  fprintf (Fp, "%-22s = 0x%06lX\n", CUSTOMCOL10, TPConfig.CustomColor10);
  fprintf (Fp, "%-22s = 0x%06lX\n", CUSTOMCOL11, TPConfig.CustomColor11);
  fprintf (Fp, "%-22s = 0x%06lX\n", CUSTOMCOL12, TPConfig.CustomColor12);
  fprintf (Fp, "%-22s = 0x%06lX\n", CUSTOMCOL13, TPConfig.CustomColor13);
  fprintf (Fp, "%-22s = 0x%06lX\n", CUSTOMCOL14, TPConfig.CustomColor14);
  fprintf (Fp, "%-22s = 0x%06lX\n", CUSTOMCOL15, TPConfig.CustomColor15);
  fclose (Fp);
}

void ChangeConfig (void)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The configuration has been changed by the user.                                                                        */
/* Import: WriteConfig, ReadConfig.                                                                                               */
/**********************************************************************************************************************************/

{
  char      BlankerVal[6];
  char      LPTNum[2];
  char      COMNum[2];
  char      PlayDevice[10];
  char     *String;
  bool      Stop         = FALSE;
  bool      AcceptConfig = FALSE;
  bool      SaveConfig   = FALSE;
  bool      Finished     = FALSE;
  bool      PathError;

  struct WindowInfo CWindow  = {125, 65, 390, 355, 390, 355, 390, 355, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "CONFIGURATION"};
  struct ButtonInfo PReload  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 50, 330, 8, DBLACK, DWHITE, NULL, 2, "@Reload", NULL};
  struct ButtonInfo PUse     = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 168, 330, 8, DBLACK, DWHITE, NULL, 22, "@Use", NULL};
  struct ButtonInfo PSave    = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 286, 330, 8, DBLACK, DWHITE, NULL, 1, "@Save", NULL};
  struct ButtonInfo CRemap   = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 162, 30, 1, DBLACK, DWHITE, NULL, 3,
                                &TPConfig.RemapBrightBlack, NULL};
  struct ButtonInfo CEmFlash = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 162, 50, 1, DBLACK, DWHITE, NULL, 23,
                                &TPConfig.EmulateFlash, NULL};
  struct ButtonInfo CSlFlash = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 162, 70, 1, DBLACK, DWHITE, NULL, 24,
                                &TPConfig.SlowFlash, NULL};
  struct ButtonInfo CCoQuit  = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 162, 90, 1, DBLACK, DWHITE, NULL, 4,
                                &TPConfig.ConfirmQuit, NULL};
  struct ButtonInfo CCoDel   = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 162, 110, 1, DBLACK, DWHITE, NULL, 5,
                                &TPConfig.ConfirmDelete, NULL};
  struct ButtonInfo CCoOver  = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 162, 130, 1, DBLACK, DWHITE, NULL, 25,
                                &TPConfig.ConfirmOverwrite, NULL};
  struct ButtonInfo CFollow  = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 162, 150, 1, DBLACK, DWHITE, NULL, 6,
                                &SystemDefaults.SyncSlider, NULL};
  struct ButtonInfo CEdInsrt = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 162, 170, 1, DBLACK, DWHITE, NULL, 20,
                                &SystemDefaults.EditInsert, NULL};
  struct ButtonInfo CAlwBck  = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 162, 190, 1, DBLACK, DWHITE, NULL, 21,
                                &TPConfig.AlwaysBackupTape, NULL};
  struct ButtonInfo CQuiet   = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 162, 210, 1, DBLACK, DWHITE, NULL, 31,
                                &SystemDefaults.Quiet, NULL};
  struct ButtonInfo CTraceSB = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 162, 230, 1, DBLACK, DWHITE, NULL, 32,
                                &TPConfig.TraceSoundBlocks, NULL};
  struct ButtonInfo CShowPlT = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 162, 250, 1, DBLACK, DWHITE, NULL, 36,
                                &TPConfig.ShowRemainingPlayTime, NULL};
  struct ButtonInfo CVOCFilt = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 162, 270, 1, DBLACK, DWHITE, NULL, 33,
                                &TPConfig.DefUseFilter, NULL};
  struct ButtonInfo CEdtPath = {BT_INPUT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT | BF_FITDIR, 214, 50, 23, DBLACK, DWHITE, NULL, 34,
                                TPConfig.EditPath, NULL};
  struct ButtonInfo CEdtPSel = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 359, 50, 1, DBLACK, DWHITE, NULL, 35, "\012", NULL};
  struct ButtonInfo CPlayDev = {BT_TEXT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 298, 100, 9, DBLACK, DWHITE, NULL, 17, NULL, NULL};
  struct ButtonInfo CPlaySel = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 359, 100, 1, DBLACK, DWHITE, NULL, 18, "\007", NULL};
  struct ButtonInfo CPlayCri = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 359, 120, 1, DBLACK, DWHITE, NULL, 19,
                                &TPConfig.PlayCriticalSection, NULL};
  struct ButtonInfo CPlayPPT = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 359, 140, 1, DBLACK, DWHITE, NULL, 37,
                                &TPConfig.ParPlayIRQsDisabled, NULL};
  struct ButtonInfo CLPTNum  = {BT_TEXT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 346, 190, 1, DBLACK, DWHITE, NULL, 13, NULL, NULL};
  struct ButtonInfo CLPTSel  = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 359, 190, 1, DBLACK, DWHITE, NULL, 14, "\007", NULL};
  struct ButtonInfo CCOMNum  = {BT_TEXT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 346, 210, 1, DBLACK, DWHITE, NULL, 15, NULL, NULL};
  struct ButtonInfo CCOMSel  = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 359, 210, 1, DBLACK, DWHITE, NULL, 16, "\007", NULL};
  struct ButtonInfo CBlankOn = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 359, 260, 1, DBLACK, DWHITE, NULL, 7,
                                &TPConfig.BlankerOn, NULL};
  struct ButtonInfo CBlankTO = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 333, 280, 5, DBLACK, DWHITE, NULL, 8,
                                NULL, NULL};
  struct ButtonInfo *ButtonP;

  CPlayDev.Contents = PlayDevice;
  CLPTNum.Contents = LPTNum;
  CCOMNum.Contents = COMNum;
  CBlankTO.Contents = BlankerVal;
  AddButton (&CWindow, &PUse, TRUE);
  AddButton (&CWindow, &PSave, FALSE);
  AddButton (&CWindow, &CRemap, FALSE);
  AddButton (&CWindow, &CEmFlash, FALSE);
  AddButton (&CWindow, &CSlFlash, FALSE);
  AddButton (&CWindow, &CCoQuit, FALSE);
  AddButton (&CWindow, &CCoDel, FALSE);
  AddButton (&CWindow, &CCoOver, FALSE);
  AddButton (&CWindow, &CFollow, FALSE);
  AddButton (&CWindow, &CEdInsrt, FALSE);
  AddButton (&CWindow, &CAlwBck, FALSE);
  AddButton (&CWindow, &CQuiet, FALSE);
  AddButton (&CWindow, &CTraceSB, FALSE);
  AddButton (&CWindow, &CShowPlT, FALSE);
  AddButton (&CWindow, &CVOCFilt, FALSE);
  AddButton (&CWindow, &CEdtPath, FALSE);
  AddButton (&CWindow, &CEdtPSel, FALSE);
  AddButton (&CWindow, &CPlayDev, FALSE);
  AddButton (&CWindow, &CPlaySel, FALSE);
  AddButton (&CWindow, &CPlayCri, FALSE);
  AddButton (&CWindow, &CPlayPPT, FALSE);
  AddButton (&CWindow, &CLPTNum, FALSE);
  AddButton (&CWindow, &CLPTSel, FALSE);
  AddButton (&CWindow, &CCOMNum, FALSE);
  AddButton (&CWindow, &CCOMSel, FALSE);
  AddButton (&CWindow, &CBlankOn, FALSE);
  AddButton (&CWindow, &CBlankTO, FALSE);
  AddButton (&CWindow, &PReload, FALSE);
  sprintf (LPTNum, "%d", TPConfig.LPTPort);
  sprintf (COMNum, "%d", TPConfig.COMPort);
  strcpy (PlayDevice, TPConfig.PlaySoundBlaster ? "Soundcard" : "Parallel");
  sprintf (BlankerVal, "%d", TPConfig.BlankerTimeOut);
  if (TPConfig.BlankerOn)
    CBlankTO.Flags &= ~BF_GHOSTED;
  else
    CBlankTO.Flags |= BF_GHOSTED;
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  WDrawShadowBox (&CWindow,  1,  0, 30, 30, "General Options");
  WDrawShadowBox (&CWindow,  1, 33,  7, 63, "Editor");
  WDrawShadowBox (&CWindow,  8, 33, 16, 63, "Play Options");
  WDrawShadowBox (&CWindow, 17, 33, 23, 63, "External Ports");
  WDrawShadowBox (&CWindow, 24, 33, 30, 63, "Blanker");
  Wprintf (&CWindow, 3,  2,  "{+1Remap Bright Black}N");
  Wprintf (&CWindow, 5,  2,  "{+1Emulate Flash}N");
  Wprintf (&CWindow, 7,  2,  "{+1Slow Flash}N");
  Wprintf (&CWindow, 9,  2,  "{+1Confirm Quit}N");
  Wprintf (&CWindow, 11, 2,  "{+1Confirm Delete}N");
  Wprintf (&CWindow, 13, 2,  "{+1Confirm Overwrite}N");
  Wprintf (&CWindow, 15, 2,  "{+1Follow Slider}N");
  Wprintf (&CWindow, 17, 2,  "{+1Edit Insert}N");
  Wprintf (&CWindow, 19, 2,  "{+1Always Backup}N");
  Wprintf (&CWindow, 21, 2,  "{+1Disable Sounds}N");
  Wprintf (&CWindow, 23, 2,  "{+1Trace Sound Blocks}N");
  Wprintf (&CWindow, 25, 2,  "{+1Show Play Time}N");
  Wprintf (&CWindow, 27, 2,  "{+1Default VOC Filter}N");
  Wprintf (&CWindow, 3,  35, "{+1Text Editor Path}N");
  Wprintf (&CWindow, 10, 35, "{+1Play Device}N");
  Wprintf (&CWindow, 12, 35, "{+1Play Windows Disabled}N");
  Wprintf (&CWindow, 14, 35, "{+1Play IRQs Disabled (Par)}N");
  Wprintf (&CWindow, 19, 35, "{+1Parallel Port     LPT}N");
  Wprintf (&CWindow, 21, 35, "{+1Serial Port       COM}N");
  Wprintf (&CWindow, 26, 35, "{+1Blanker Installed}N");
  Wprintf (&CWindow, 28, 35, "{+1Blanker Time-Out}N");
  while (!Finished)
  {
    while (!AcceptConfig)
    {
      Stop = FALSE;
      while (!Stop)
      {
        switch (HandleWindow (&CWindow, &ButtonP))
        {
          case WINH_BUTTBOOL: switch (ButtonP->ReactNum)
                              {
                                case 7  : if (TPConfig.BlankerOn)                                                   /* Blanker on */
                                            CBlankTO.Flags &= ~BF_GHOSTED;
                                          else
                                            CBlankTO.Flags |= BF_GHOSTED;
                                          DrawButton (&CWindow, &CBlankTO);
                                          if (TPConfig.BlankerOn)
                                            InstallBlanker (TPConfig.BlankerTimeOut, NULL);
                                          else
                                            DisableBlanker ();
                                          break;
                                case 21 : if (!TapeWindow[TWLEFT].NonexistingTape)
                                            TapeWindow[TWLEFT].BackupTape = TPConfig.AlwaysBackupTape;
                                          if (!TapeWindow[TWRIGHT].NonexistingTape)
                                            TapeWindow[TWRIGHT].BackupTape = TPConfig.AlwaysBackupTape;
                                          break;
                              }
                              break;
          case WINH_CLOSE   : Stop = TRUE;
                              Finished = TRUE;
                              break;
          case WINH_RNOHAND : switch (ButtonP->ReactNum)
                              {
                                case 1  : Stop = TRUE; SaveConfig = TRUE; break;                                          /* Save */
                                case 22 : Stop = TRUE; Finished = TRUE; break;                                             /* Use */
                                case 2  : ReadConfig ();
                                          sprintf (BlankerVal, "%d", TPConfig.BlankerTimeOut);
                                          if (TPConfig.BlankerOn)
                                            CBlankTO.Flags &= ~BF_GHOSTED;
                                          else
                                            CBlankTO.Flags |= BF_GHOSTED;
                                          DrawButton (&CWindow, &CRemap);
                                          DrawButton (&CWindow, &CEmFlash);
                                          DrawButton (&CWindow, &CSlFlash);
                                          DrawButton (&CWindow, &CCoQuit);
                                          DrawButton (&CWindow, &CCoDel);
                                          DrawButton (&CWindow, &CCoOver);
                                          DrawButton (&CWindow, &CFollow);
                                          DrawButton (&CWindow, &CEdInsrt);
                                          DrawButton (&CWindow, &CAlwBck);
                                          DrawButton (&CWindow, &CQuiet);
                                          DrawButton (&CWindow, &CBlankOn);
                                          UpdateButtonText (&CWindow, &CBlankTO, BF_FITTXT);
                                          break;
                                case 14 : TPConfig.LPTPort ++;                                                         /* LPT Nr. */
                                          while (TPConfig.LPTPort < 4 && TPConfig.LPTAddr[TPConfig.LPTPort - 1] == 0x0000)
                                            TPConfig.LPTPort ++;                                     /* (Skip non-existing ports) */
                                          if (TPConfig.LPTPort == 4)
                                            TPConfig.LPTPort = 1;
                                          sprintf (LPTNum, "%d", TPConfig.LPTPort);
                                          DrawButton (&CWindow, &CLPTNum);
                                          break;
                                case 16 : TPConfig.COMPort ++;                                                         /* COM Nr. */
                                          while (TPConfig.COMPort < 5 && TPConfig.COMAddr[TPConfig.COMPort - 1] == 0x0000)
                                            TPConfig.COMPort ++;                                     /* (Skip non-existing ports) */
                                          if (TPConfig.COMPort == 5)
                                            TPConfig.COMPort = 1;
                                          sprintf (COMNum, "%d", TPConfig.COMPort);
                                          DrawButton (&CWindow, &CCOMNum);
                                          break;
                                case 18 : TPConfig.PlaySoundBlaster = !TPConfig.PlaySoundBlaster;                  /* Play Device */
                                          strcpy (PlayDevice, TPConfig.PlaySoundBlaster ? "Soundcard" : "Parallel");
                                          DrawButton (&CWindow, &CPlayDev);
                                          break;
                                case 35 : _GetNewFileName ("Select game info editor", CEdtPath.Contents);
                                          UpdateButtonText (&CWindow, &CEdtPath, BF_FITDIR);
                                          break;
                              }
                              break;
          case WINH_EDITEND : if (ButtonP->Flags & BF_FITINPUT)                                        /* Contents were changed ? */
                                switch (ButtonP->ReactNum)
                                {
                                  case 8  : TPConfig.BlankerTimeOut = Blanker.TimeOut = atoi (BlankerVal);
                                            sprintf (BlankerVal, "%d", TPConfig.BlankerTimeOut);
                                            UpdateButtonText (&CWindow, &CBlankTO, BF_FITTXT);
                                            break;
                                }
        }
        if (GlobalKey == K_ESC)
          Stop = Finished = TRUE;
        MouseStatus ();
      }
      String = TestName (TPConfig.EditPath, FALSE, &PathError);
      if (PathError)
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR: %s", String);
      else
      {
        strcpy (TPConfig.EditPath, String);
        AcceptConfig = TRUE;
      }
    }
    if (SaveConfig)
    {
      WriteConfig ();
      AcceptConfig = SaveConfig = FALSE;
    }
  }
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
}

void AboutBox (bool MakeWindow)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : An `about'-box has been shown on the screen and closed by the user.                                                    */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct ButtonInfo  POk      = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 79, 215, 8, DBLACK, DWHITE, NULL, 1, "@Ok", NULL};
  struct ButtonInfo *ButtonP;
  short              HandleRet;

  memset (&AboutWindow, 0, sizeof (struct WindowInfo));
  //AboutWindow.StartX      = MakeWindow ? 186 : 215;
  AboutWindow.StartX      = 215;
  AboutWindow.StartY      = MakeWindow ? 144 : 10;
  AboutWindow.Width       = AboutWindow.MinWidth = AboutWindow.MaxWidth = 210;
  AboutWindow.Height      = AboutWindow.MinHeight = AboutWindow.MaxHeight = MakeWindow ? 240 : 215;
  AboutWindow.BorderSize  = MakeWindow ? 1 : 3;
  AboutWindow.BorderFlags = MakeWindow ? (WINB_OUT | WINB_SURROUND) : WINB_NONE;
  AboutWindow.APen        = DBLACK;
  AboutWindow.BPen        = DWHITE;
  AboutWindow.Font        = NULL;
  AboutWindow.Flags       = MakeWindow ? (WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER) : WINF_XMSBUFFER;
  AboutWindow.HeaderFont  = NULL;
  AboutWindow.Header      = "About This Program";
  if (MakeWindow)
    AddButton (&AboutWindow, &POk, TRUE);
  DrawWindow (&AboutWindow, NULL, WING_CLOSE, TRUE);
  WDrawPicture (&AboutWindow, 40, 7, SpectrumLogo);
  Wprintf (&AboutWindow,  4, TEXT_CENTER,          "{+3Taper}N");
  Wprintf (&AboutWindow,  4, TEXT_MCENTER,         "{+1Taper}N");
  Wprintf (&AboutWindow,  7, TEXT_CENTER,              "By");
  Wprintf (&AboutWindow,  8, TEXT_CENTER,    "{TMartijn van der Heide}T");
  Wprintf (&AboutWindow, 10, TEXT_CENTER,     "Version 2.07 - 09/12/01");
  Wprintf (&AboutWindow, 11, TEXT_CENTER,         "32-bit version");
  Wprintf (&AboutWindow, 13, TEXT_CENTER,     "Copyleft (c) 1997-2001");
  Wprintf (&AboutWindow, 14, TEXT_CENTER,   "ThunderWare Research Center");
  Wprintf (&AboutWindow, 16, TEXT_CENTER,  "This program is free software");
  Wprintf (&AboutWindow, 18, TEXT_CENTER,   "See 'COPYING' for more info");
  if (!MakeWindow)
  {
    WDrawBox (&AboutWindow, FALSE, DBLACK, 0, 0, AboutWindow.Width - 1, AboutWindow.Height - 1);
    WDrawBox (&AboutWindow, FALSE, DBLACK, 1, 1, AboutWindow.Width - 2, AboutWindow.Height - 2);
    return;
  }
  while ((HandleRet = HandleWindow (&AboutWindow, &ButtonP)) != WINH_RNOHAND && HandleRet != WINH_CLOSE && GlobalKey != K_ESC)
    MouseStatus ();
  DestroyWindow (&AboutWindow);
  GlobalKey = 0x0000;
}

void SysInfoBox (void)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : A `system info'-box has been shown on the screen and closed by the user.                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct WindowInfo CWindow  = {212, 100, 216, 265, 216, 265, 216, 265, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "System Information"};
  struct ButtonInfo POk      = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 81, 240, 8, DBLACK, DWHITE, NULL, 1, "@Ok", NULL};
  struct ButtonInfo     *ButtonP;
  short                  HandleRet;
  short                  LineAdd     = 0;
  short                  CurLine     = 1;
  byte                  *Version;
  char                  *MouseType[] = { "Unknown", "Bus", "Serial", "InPort", "PS/2", "HP" };
  char                   MSWinMode[6];
  char                   MSWinVersion[8];

  if (MSWindowsRunning)
    LineAdd ++;
  if (SoundBlasterInfo.Detected)                                                                  /* Did we find a SoundBlaster ? */
    LineAdd += 4;
  CWindow.StartY -= (LineAdd * (CurrentFont->Height + 2)) / 2;                                   /* Keep the window in the middle */
  CWindow.Height = CWindow.MinHeight = CWindow.MaxHeight = CWindow.Height + LineAdd * (CurrentFont->Height + 2);
  POk.StartY += LineAdd * (CurrentFont->Height + 2);
  AddButton (&CWindow, &POk, TRUE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  Wprintf (&CWindow,  CurLine ++, 1, "{+1OS}N       DOS Version : %d.%02d", _osmajor, _osminor);
  Wprintf (&CWindow,  CurLine ++, 1, "         WB Version  : %d.%02d.%d", WBMajorVersion, WBMinorVersion, WBRevision);
  if (MSWindowsRunning)
  {
    Version = (byte *)&MSWindowsVersion;
    if (MSWindowsStopped)
      strcpy (MSWinMode, "(Dis)");
    else
      strcpy (MSWinMode, MSWindowsMode == 0x0002 ? "(Std)" : "(Enh)");
    if (MSWindowsVersion >= 0x0400)                                                       /* Windows95 is reported as version 4.0 */
    {
      if (MSWindowsVersion == 0xFFFF)                                                        /* The special value for WindowsNT ? */
        strcpy (MSWinVersion, "WinNT");
      else if (MSWindowsVersion < 0x0403)
        strcpy (MSWinVersion, "Win95");
      else if (MSWindowsVersion < 0x0500)
        strcpy (MSWinVersion, "Win96");                                                                    /* A.k.a. OSR2 (v4.3+) */
      else
        strcpy (MSWinVersion, "Win98");                                       /* Assume Windows98 will be reported as version 5.0 */
      if (!MSWindowsStopped)
        MSWinMode[0] = '\0';                                                   /* (There's no such thing as Std or Enh Windows95) */
    }
    else
      sprintf (MSWinVersion, "%d.%d", *(Version + 1), *Version);
    Wprintf (&CWindow, CurLine ++, 1, "         MS Windows  : %s %s", MSWinVersion, MSWinMode);
  }
  CurLine ++;
  Wprintf (&CWindow, CurLine ++, 1, "{+1DISPLAY}N  Color       : %s", Screen.ColorDisplay ? "Yes" : "No");
  Wprintf (&CWindow, CurLine ++, 1, "         Start       : %05lX", Screen.DisplayStart);
  if (Screen.UseRIL)
    Wprintf (&CWindow, CurLine ++, 1, "         RIL Version : %d.%02d", Screen.RILMajorVersion, Screen.RILMinorVersion);
  else
    Wprintf (&CWindow, CurLine ++, 1, "         Control     : Direct");
  CurLine ++;
  Wprintf (&CWindow, CurLine ++, 1, "{+1KEYBOARD}N Enhanced    : %s", EnhancedKeyboard ? "Yes" : "No");
  Version = (byte *)&(Mouse.DriverVersion);
  CurLine ++;
  Wprintf (&CWindow, CurLine ++, 1, "{+1MOUSE}N    Version     : %X.%02X", *(Version + 1), *Version);
  Wprintf (&CWindow, CurLine ++, 1, "         Buttons     : %d", Mouse.NumButtons);
  Wprintf (&CWindow, CurLine ++, 1, "         Type        : %s", MouseType[Mouse.MouseType]);
  if (Mouse.MouseIRQ == 0 || Mouse.MouseIRQ == 255)                                            /* IRQ 0 is reserved for PS/2 mice */
    Wprintf (&CWindow, CurLine ++, 1, "         IRQ         : (PS/2)");
  else
    Wprintf (&CWindow, CurLine ++, 1, "         IRQ         : %d", Mouse.MouseIRQ);
  CurLine ++;
  if (SoundBlasterInfo.Detected)
  {
    Wprintf (&CWindow, CurLine ++, 1, "{+1SOUND}N    BaseAddress : %03X", SoundBlasterInfo.BaseAddress);
    Wprintf (&CWindow, CurLine ++, 1, "         IRQ         : %d", SoundBlasterInfo.IRQ);
    Wprintf (&CWindow, CurLine ++, 1, "         DMA Channel : %d", SoundBlasterInfo.DMAChannel);
    switch (SoundBlasterInfo.DSPMajorVersion)
    {
      case 1 :
      case 2 : sprintf (MSWinVersion, "SB %u.%u", SoundBlasterInfo.DSPMajorVersion, SoundBlasterInfo.DSPMinorVersion); break;
      case 3 : strcpy (MSWinVersion, SoundBlasterInfo.DSPMinorVersion == 0 ? "SBPro" : "SBPro2"); break;
      case 4 : if (SoundBlasterInfo.DSPMinorVersion < 12)       strcpy (MSWinVersion, "SB16");
               else if (SoundBlasterInfo.DSPMinorVersion < 16)  strcpy (MSWinVersion, "AWE32");
               else                                             strcpy (MSWinVersion, "AWE64+");
               break;
      default: strcpy (MSWinVersion, "Dunno"); break;
    }
    Wprintf (&CWindow, CurLine ++, 1, "         Model       : %s", MSWinVersion);
    switch (SoundBlasterInfo.MixerType)
    {
      case SBMTYPE_NONE   : strcpy (MSWinVersion, "None"); break;
      case SBMTYPE_SBPRO  : strcpy (MSWinVersion, "SBPro"); break;
      case SBMTYPE_SB16   : strcpy (MSWinVersion, "SB16"); break;
      case SBMTYPE_AD1848 : strcpy (MSWinVersion, "AD1848"); break;
      default             : strcpy (MSWinVersion, "Dunno"); break;
    }
    Wprintf (&CWindow, CurLine ++, 1, "         Soft Mixer  : %s", MSWinVersion);
  }
  else
    Wprintf (&CWindow, CurLine ++, 1, "{+1SOUND}N    Device      : None");
  CurLine ++;
  if (!DPMIServerDetected)
    Wprintf (&CWindow, CurLine ++, 1, "{+1DPMI}N                 : None");
  else
  {
    Wprintf (&CWindow, CurLine ++, 1, "{+1DPMI}N     Version     : %d.%d", DPMIMajorVersion, DPMIMinorVersion);
    Wprintf (&CWindow, CurLine ++, 1, "         Int Mode    : %s", DPMIFlags & 0x02 ? "Real" : "V86");
    Wprintf (&CWindow, CurLine ++, 1, "         VMEM Support: %s", DPMIFlags & 0x04 ? "Yes" : "No");
    //Wprintf (&CWindow, CurLine ++, 1, "         Processor   : %d", ProcessorType);
  }
  CurLine ++;
  Wprintf (&CWindow, CurLine ++, 1, "%s     Total left  : %lu", DPMIFlags & 0x04 ? "{+1VMEM}N" : "{+1MEM}N ", _memavl);
  while ((HandleRet = HandleWindow (&CWindow, &ButtonP)) != WINH_RNOHAND && HandleRet != WINH_CLOSE && GlobalKey != K_ESC)
    MouseStatus ();
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
}
