/**********************************************************************************************************************************/
/* Project      : TAPER                                                                                                           */
/* Module       : TPFILE.C                                                                                                        */
/* Description  : File handler                                                                                                    */
/* Version type : Program module                                                                                                  */
/* Last changed : 17-05-1999  14:25                                                                                               */
/* Update count : 7                                                                                                               */
/* OS type      : PC (DOS)                                                                                                        */
/* Compiler     : Watcom C32 Optimizing Compiler Version 10.x                                                                     */
/* Linker       : Watcom C Linker                                                                                                 */
/*                                                                                                                                */
/*                Copyleft (C) 1997-2001 ThunderWare Research Center, written by Martijn van der Heide.                           */
/*                                                                                                                                */
/*                This program is free software; you can redistribute it and/or                                                   */
/*                modify it under the terms of the GNU General Public License                                                     */
/*                as published by the Free Software Foundation; either version 2                                                  */
/*                of the License, or (at your option) any later version.                                                          */
/*                                                                                                                                */
/*                This program is distributed in the hope that it will be useful,                                                 */
/*                but WITHOUT ANY WARRANTY; without even the implied warranty of                                                  */
/*                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                                   */
/*                GNU General Public License for more details.                                                                    */
/*                                                                                                                                */
/*                You should have received a copy of the GNU General Public License                                               */
/*                along with this program; if not, write to the Free Software                                                     */
/*                Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                     */
/*                                                                                                                                */
/**********************************************************************************************************************************/

#define __TPFILE_MAIN__

#include "wbench32/wbench32.h"
#include <dos.h>
#include <fcntl.h>
#include <io.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>
#include "tpmain.h"
#include "tpfile.h"

#pragma pack                ()

#include "specudg.inc"                                                                         /* Include Spectrum Font UDG remap */

#define HDRLEN               19                                                      /* A header is 17 bytes + Flag byte + Parity */

/**********************************************************************************************************************************/
/* Define the global variables                                                                                                    */
/**********************************************************************************************************************************/

struct RemapCharacterSet_s TokenMap[256] = {
   {"\x3F", 0x00 }, {"\x3F", 0x00  }, {"\x3F", 0x00   }, {"\x3F", 0x00 },  {"\x3F", 0x00 }, {"\x3F", 0x00 },
   {"\t",   0x00 }, {"\x3F", 0x00  }, {"\x3F", 0x00   }, {"\x3F", 0x00 },  {"\x3F", 0x00 }, {"\x3F", 0x00 },
   {"\x3F", 0x00 }, {"\r",   0x00  }, {"\0",   0x00   }, {"\x3F", 0x00 },  {"\0", F_INK  }, {"\0", F_PAPER},
   {"\0", F_FLASH}, {"\0", F_BRIGHT}, {"\0", F_INVERSE}, {"\x3F", F_OVER}, {"\0", F_SKIP2}, {"\0", F_SKIP2},
   {"\x3F", 0x00 }, {"\x3F", 0x00  }, {"\x3F", 0x00   }, {"\x3F", 0x00 },  {"\x3F", 0x00 }, {"\x3F", 0x00 },
   {"\x3F", 0x00 }, {"\x3F", 0x00  },

   {"\x20", 0x00}, {"\x21", 0x00}, {"\x22", 0x00}, {"\x23", 0x00}, {"\x24", 0x00}, {"\x25", 0x00}, {"\x26", 0x00}, {"\x27", 0x00},
   {"\x28", 0x00}, {"\x29", 0x00}, {"\x2A", 0x00}, {"\x2B", 0x00}, {"\x2C", 0x00}, {"\x2D", 0x00}, {"\x2E", 0x00}, {"\x2F", 0x00},
   {"\x30", 0x00}, {"\x31", 0x00}, {"\x32", 0x00}, {"\x33", 0x00}, {"\x34", 0x00}, {"\x35", 0x00}, {"\x36", 0x00}, {"\x37", 0x00},
   {"\x38", 0x00}, {"\x39", 0x00}, {"\x3A", 0x00}, {"\x3B", 0x00}, {"\x3C", 0x00}, {"\x3D", 0x00}, {"\x3E", 0x00}, {"\x3F", 0x00},
   {"\x40", 0x00}, {"\x41", 0x00}, {"\x42", 0x00}, {"\x43", 0x00}, {"\x44", 0x00}, {"\x45", 0x00}, {"\x46", 0x00}, {"\x47", 0x00},
   {"\x48", 0x00}, {"\x49", 0x00}, {"\x4A", 0x00}, {"\x4B", 0x00}, {"\x4C", 0x00}, {"\x4D", 0x00}, {"\x4E", 0x00}, {"\x4F", 0x00},
   {"\x50", 0x00}, {"\x51", 0x00}, {"\x52", 0x00}, {"\x53", 0x00}, {"\x54", 0x00}, {"\x55", 0x00}, {"\x56", 0x00}, {"\x57", 0x00},
   {"\x58", 0x00}, {"\x59", 0x00}, {"\x5A", 0x00}, {"\x5B", 0x00}, {"\x5C", 0x00}, {"\x5D", 0x00}, {"\x5E", 0x00}, {"\x5F", 0x00},
   {"\x60", 0x00}, {"\x61", 0x00}, {"\x62", 0x00}, {"\x63", 0x00}, {"\x64", 0x00}, {"\x65", 0x00}, {"\x66", 0x00}, {"\x67", 0x00},
   {"\x68", 0x00}, {"\x69", 0x00}, {"\x6A", 0x00}, {"\x6B", 0x00}, {"\x6C", 0x00}, {"\x6D", 0x00}, {"\x6E", 0x00}, {"\x6F", 0x00},
   {"\x70", 0x00}, {"\x71", 0x00}, {"\x72", 0x00}, {"\x73", 0x00}, {"\x74", 0x00}, {"\x75", 0x00}, {"\x76", 0x00}, {"\x77", 0x00},
   {"\x78", 0x00}, {"\x79", 0x00}, {"\x7A", 0x00}, {"\x7B", 0x00}, {"\x7C", 0x00}, {"\x7D", 0x00}, {"\x7E", 0x00}, {"\x7F", 0x00},

   {"\x80", 0x00}, {"\x81", 0x00}, {"\x82", 0x00}, {"\x83", 0x00}, {"\x84", 0x00}, {"\x85", 0x00}, {"\x86", 0x00}, {"\x87", 0x00},
   {"\x88", 0x00}, {"\x89", 0x00}, {"\x8A", 0x00}, {"\x8B", 0x00}, {"\x8C", 0x00}, {"\x8D", 0x00}, {"\x8E", 0x00}, {"\x8F", 0x00},

   {"\x90", 0x00}, {"\x91", 0x00}, {"\x92", 0x00}, {"\x93", 0x00}, {"\x94", 0x00}, {"\x95", 0x00}, {"\x96", 0x00}, {"\x97", 0x00},
   {"\x98", 0x00}, {"\x99", 0x00}, {"\x9A", 0x00}, {"\x9B", 0x00}, {"\x9C", 0x00}, {"\x9D", 0x00}, {"\x9E", 0x00}, {"\x9F", 0x00},
   {"\xA0", 0x00}, {"\xA1", 0x00}, {"\xA2", 0x00},

   {"SPECTRUM",  F_PAD_R | F_PAD_L}, {"PLAY",    F_PAD_R | F_PAD_L},               /* For Spectrum 128. 48 are remapped (shorter) */

   {"RND",       0x00             }, {"INKEY$",  0x00             }, {"PI",     0x00             }, {"FN",       F_PAD_R          },
   {"POINT",     F_PAD_R          }, {"SCREEN$", F_PAD_R          }, {"ATTR",   F_PAD_R          }, {"AT",       F_PAD_R          },
   {"TAB",       F_PAD_R          }, {"VAL$",    F_PAD_R          }, {"CODE",   F_PAD_R          }, {"VAL",      F_PAD_R          },
   {"LEN",       F_PAD_R          }, {"SIN",     F_PAD_R          }, {"COS",    F_PAD_R          }, {"TAN",      F_PAD_R          },
   {"ASN",       F_PAD_R          }, {"ACS",     F_PAD_R          }, {"ATN",    F_PAD_R          }, {"LN",       F_PAD_R          },
   {"EXP",       F_PAD_R          }, {"INT",     F_PAD_R          }, {"SQR",    F_PAD_R          }, {"SGN",      F_PAD_R          },
   {"ABS",       F_PAD_R          }, {"PEEK",    F_PAD_R          }, {"IN",     F_PAD_R          }, {"USR",      F_PAD_R          },
   {"STR$",      F_PAD_R          }, {"CHR$",    F_PAD_R          }, {"NOT",    F_PAD_R          }, {"BIN",      F_PAD_R          },
   {"OR",        F_PAD_R | F_PAD_L}, {"AND",     F_PAD_R | F_PAD_L}, {"<=",     0x00             }, {">=",       0x00             },
   {"<>",        0x00             }, {"LINE",    F_PAD_R          }, {"THEN",   F_PAD_R | F_PAD_L}, {"TO",       F_PAD_R | F_PAD_L},
   {"STEP",      F_PAD_R | F_PAD_L}, {"DEF FN",  F_PAD_R | F_PAD_L}, {"CAT",    F_PAD_R | F_PAD_L}, {"FORMAT",   F_PAD_R | F_PAD_L},
   {"MOVE",      F_PAD_R | F_PAD_L}, {"ERASE",   F_PAD_R | F_PAD_L}, {"OPEN #", F_PAD_R | F_PAD_L}, {"CLOSE #",  F_PAD_R | F_PAD_L},
   {"MERGE",     F_PAD_R | F_PAD_L}, {"VERIFY",  F_PAD_R | F_PAD_L}, {"BEEP",   F_PAD_R | F_PAD_L}, {"CIRCLE",   F_PAD_R | F_PAD_L},
   {"INK",       F_PAD_R | F_PAD_L}, {"PAPER",   F_PAD_R | F_PAD_L}, {"FLASH",  F_PAD_R | F_PAD_L}, {"BRIGHT",   F_PAD_R | F_PAD_L},
   {"INVERSE",   F_PAD_R | F_PAD_L}, {"OVER",    F_PAD_R | F_PAD_L}, {"OUT",    F_PAD_R | F_PAD_L}, {"LPRINT",   F_PAD_R | F_PAD_L},
   {"LLIST",     F_PAD_R | F_PAD_L}, {"STOP",    F_PAD_R | F_PAD_L}, {"READ",   F_PAD_R | F_PAD_L}, {"DATA",     F_PAD_R | F_PAD_L},
   {"RESTORE",   F_PAD_R | F_PAD_L}, {"NEW",     F_PAD_R | F_PAD_L}, {"BORDER", F_PAD_R | F_PAD_L}, {"CONTINUE", F_PAD_R | F_PAD_L},
   {"DIM",       F_PAD_R | F_PAD_L}, {"REM",     F_PAD_R | F_PAD_L}, {"FOR",    F_PAD_R | F_PAD_L}, {"GO TO",    F_PAD_R | F_PAD_L},
   {"GO SUB",    F_PAD_R | F_PAD_L}, {"INPUT",   F_PAD_R | F_PAD_L}, {"LOAD",   F_PAD_R | F_PAD_L}, {"LIST",     F_PAD_R | F_PAD_L},
   {"LET",       F_PAD_R | F_PAD_L}, {"PAUSE",   F_PAD_R | F_PAD_L}, {"NEXT",   F_PAD_R | F_PAD_L}, {"POKE",     F_PAD_R | F_PAD_L},
   {"PRINT",     F_PAD_R | F_PAD_L}, {"PLOT",    F_PAD_R | F_PAD_L}, {"RUN",    F_PAD_R | F_PAD_L}, {"SAVE",     F_PAD_R | F_PAD_L},
   {"RANDOMIZE", F_PAD_R | F_PAD_L}, {"IF",      F_PAD_R | F_PAD_L}, {"CLS",    F_PAD_R | F_PAD_L}, {"DRAW",     F_PAD_R | F_PAD_L},
   {"CLEAR",     F_PAD_R | F_PAD_L}, {"RETURN",  F_PAD_R | F_PAD_L}, {"COPY",   F_PAD_R | F_PAD_L}};

/**********************************************************************************************************************************/
/* Define the static variables                                                                                                    */
/**********************************************************************************************************************************/

static bool      _PadLeft                = FALSE;
static short     _SpectrumColorTable[8] = { 0, 1, 4, 5, 2, 3, 6, 7 };

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE EXPORTED LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

short RemapSpectrumString (byte *SpectrumString, short Length, char *RemappedString, short MaxLength, bool DropColors,
                           bool InitialPad, short *RawLength, bool ToCombo, bool RealNumbers, bool DropTABs)

/**********************************************************************************************************************************/
/* Pre   : `SpectrumString' holds the string to be remapped, the length is `Length', The result should go to `RemappedString',    */
/*         which can contain a maximum of `MaxLength' characters.                                                                 */
/*         If `DropColors' is TRUE, no color attributes are to be inserted.                                                       */
/*         If `InitialPad' is TRUE, the first character inserted is a (padding) space (needed for listings).                      */
/*         If `RealNumbers' is TRUE, the stored 5-byte floating point form will be used in stead of the listed values.            */
/*         The allocated length of `RemappedString' should be at least 9 bytes larger then `MaxLength'!                           */
/* Post  : The string has been remapped to the full Spectrum equivalent. All tokens are expanded. The return value is the number  */
/*         of bytes in the `RemappedString'.                                                                                      */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  short register  CharCnt;
  short register  StrLength  = 0;
  short register  SkipLength = 0;
  short register  RealLength = 0;                                                                  /* (Length without attributes) */
  short register  Cnt;
  bool  register  Done = FALSE;
  static char     Buffer[20];
  double          FPValue;
  double          Power;
  double          Raised;
  double          Result;

  RemappedString[0] = '\0';
  _PadLeft = InitialPad;
  for (CharCnt = 0 ; CharCnt < Length ; CharCnt ++)
  {
    Done = FALSE;
    if (RemapCharacterSet[SpectrumString[CharCnt]].Flags & F_PAD_L)                               /* This char has left padding ? */
      _PadLeft = TRUE;
    if (_PadLeft)
    {
      strcat (RemappedString + StrLength ++, " ");
      RealLength ++;
    }
    if (RemapCharacterSet[SpectrumString[CharCnt]].Flags & F_PAD_R)                              /* This char has right padding ? */
      _PadLeft = TRUE;                                                                                  /* Prepare for next token */
    else
      _PadLeft = FALSE;
    if (ToCombo)
      switch (RemapCharacterSet[SpectrumString[CharCnt]].Flags)
      {
        case F_PAPER   : if (!DropColors)
                         {
                           sprintf (RemappedString + StrLength, "{-%d", _SpectrumColorTable[SpectrumString[CharCnt + 1] & 0x07]);
                           StrLength += 3;
                         }
                         SkipLength = 1;
                         break;
        case F_INK     : if (!DropColors)
                         {
                           sprintf (RemappedString + StrLength, "{+%d", _SpectrumColorTable[SpectrumString[CharCnt + 1] & 0x07]);
                           StrLength += 3;
                         }
                         SkipLength = 1;
                         break;
        case F_BRIGHT  : if (!DropColors)
                         {
                           if (SpectrumString[CharCnt + 1] == 0)
                             strcat (RemappedString + StrLength, "{+/{-/");
                           else
                             strcat (RemappedString + StrLength, "{+*{-*");
                           StrLength += 6;
                         }
                         SkipLength = 1;
                         break;
        case F_INVERSE : if (!DropColors)
                         {
                           if (SpectrumString[CharCnt + 1] == 0)
                             strcat (RemappedString + StrLength, "}I");
                           else
                             strcat (RemappedString + StrLength, "{I");
                           StrLength += 2;
                         }
                         SkipLength = 1;
                         break;
        case F_FLASH   : if (!DropColors)
                         {
                           if (SpectrumString[CharCnt + 1] == 0)
                           {
                             strcat (RemappedString + StrLength, "}L");
                             StrLength += 2;
                           }
                           else
                           {
                             strcat (RemappedString + StrLength, "{L2");
                             StrLength += 3;
                           }
                         }
                         SkipLength = 1;
                         break;
        case F_SKIP2   : SkipLength = 2;
                         break;
        default        : if (!strcmp (RemapCharacterSet[SpectrumString[CharCnt]].Token, "{"))
                         {
                           strcat (RemappedString + StrLength, "{");                                   /* (Insert extra switches) */
                           StrLength += 1;
                         }
                         else if (!strcmp (RemapCharacterSet[SpectrumString[CharCnt]].Token, "}"))
                         {
                           strcat (RemappedString + StrLength, "}");
                           StrLength += 1;
                         }
                         SkipLength = 0;
      }
    else                                                                                                   /* (Writing to a file) */
    {
      SkipLength = 0;
      if (SpectrumString[CharCnt] < 0x20)
      {
        Buffer[0] = '\0';
        switch (RemapCharacterSet[SpectrumString[CharCnt]].Flags)
        {
          case F_PAPER   : if (!DropColors)
                             sprintf (Buffer, "{PAPER %d}", SpectrumString[CharCnt + 1]);
                           SkipLength = 1;
                           break;
          case F_INK     : if (!DropColors)
                             sprintf (Buffer, "{INK %d}", SpectrumString[CharCnt + 1]);
                           SkipLength = 1;
                           break;
          case F_BRIGHT  : if (!DropColors)
                             sprintf (Buffer, "{BRIGHT %d}", SpectrumString[CharCnt + 1]);
                           SkipLength = 1;
                           break;
          case F_INVERSE : if (!DropColors)
                             sprintf (Buffer, "{INVERSE %d}", SpectrumString[CharCnt + 1]);
                           SkipLength = 1;
                           break;
          case F_FLASH   : if (!DropColors)
                             sprintf (Buffer, "{FLASH %d}", SpectrumString[CharCnt + 1]);
                           SkipLength = 1;
                           break;
          case F_OVER    : if (!DropColors)
                             sprintf (Buffer, "{OVER %d}", SpectrumString[CharCnt + 1]);
                           SkipLength = 1;
                           break;
          case F_SKIP2   : if (!DropColors)
                             sprintf (Buffer, "{%s %d,%d}", SpectrumString[CharCnt] == 0x16 ? "AT" : "TAB",
                                    SpectrumString[CharCnt + 1], SpectrumString[CharCnt + 2]);
                           SkipLength = 2;
                           break;
          default        : if (SpectrumString[CharCnt] == 0x06)
                             strcpy (Buffer, "\t");
                           else if (SpectrumString[CharCnt] == 0x0D)
                             strcpy (Buffer, "\n");
                           else if (SpectrumString[CharCnt] != 0x0E)
                             sprintf (Buffer, "{%02X}", SpectrumString[CharCnt]);
                           break;
        }
        Done = TRUE;
      }
      else if (SpectrumString[CharCnt] >= 0x80 && SpectrumString[CharCnt] < 0x88)
      {
        if (SpectrumString[CharCnt] == 0x80)
          strcpy (Buffer, "{-8}");
        else
          sprintf (Buffer, "{-%X}", SpectrumString[CharCnt] - 0x80);
        Done = TRUE;
      }
      else if (SpectrumString[CharCnt] >= 0x88 && SpectrumString[CharCnt] < 0x90)
      {
        if (SpectrumString[CharCnt] == 0x8F)
          strcpy (Buffer, "{+8}");
        else
          sprintf (Buffer, "{+%X}", 7 - (SpectrumString[CharCnt] - 0x88));
        Done = TRUE;
      }
      else if (SpectrumString[CharCnt] >= 0x90 && SpectrumString[CharCnt] < 0xA3)
      {
        sprintf (Buffer, "{%c}", SpectrumString[CharCnt] - 0x90 + 'A');
        Done = TRUE;
      }
      else if (SpectrumString[CharCnt] >= 0xA3 && SpectrumString[CharCnt] < 0xA4)
        if (!strcmp (RemapCharacterSet[0xA3].Token, "SPECTRUM"))                              /* Special exception - doing 128K ? */
        {
          strcpy (Buffer, RemapCharacterSet[SpectrumString[CharCnt]].Token);
          Done = TRUE;
        }
        else
        {
          sprintf (Buffer, "{%c}", SpectrumString[CharCnt] - 0x90 + 'A');
          Done = TRUE;
        }
      else if (SpectrumString[CharCnt] == 0x7F)                                             /* Special exception - copyright sign */
      {
        strcpy (Buffer, "{(C)}");
        Done = TRUE;
      }
    }
    if (StrLength >= MaxLength)
      return (MaxLength);
    if (!Done && RealNumbers)
      if ((isdigit (SpectrumString[CharCnt]) || SpectrumString[CharCnt] == '.') && !isalpha (SpectrumString[CharCnt - 1]))
      {                                                                                                    /* Start of a number ? */
        Cnt = CharCnt + 1;
        while (isdigit (SpectrumString[Cnt]))                                                   /* This is the format of a number */
          Cnt ++;
        if (SpectrumString[Cnt] == '.')
        {
          Cnt ++;
          while (isdigit (SpectrumString[Cnt]))
            Cnt ++;
        }
        if (toupper (SpectrumString[Cnt]) == 'E')
        {
          Cnt ++;
          if (SpectrumString[Cnt] == '+' || SpectrumString[Cnt] == '-')
            Cnt ++;
          while (isdigit (SpectrumString[Cnt]))
            Cnt ++;
        }
        if (SpectrumString[Cnt ++] == 0x0E)                                                                    /* Number marker ? */
        {
          if (SpectrumString[Cnt] == 0x00)                                                                /* Small integer form ? */
          {
            FPValue = (double)SpectrumString[Cnt + 2] + 256 * (double)SpectrumString[Cnt + 3];
            if (SpectrumString[Cnt + 1])                                                                            /* Negative ? */
              FPValue -= (double)65536;                                                         /* (Maintain bug in Spectrum ROM) */
          }
          else
          {
            Power = pow (2, ((double)SpectrumString[Cnt] - 129));
            if (SpectrumString[Cnt + 1] & 0x80)                                                     /* Determine negative numbers */
              Power *= -1;
            Raised = 128;
            Result = (SpectrumString[Cnt + 1] & 0x7F) / Raised;
            Raised *= 256;
            Result += SpectrumString[Cnt + 2] / Raised;
            Raised *= 256;
            Result += SpectrumString[Cnt + 3] / Raised;
            Raised *= 256;
            Result += SpectrumString[Cnt + 4] / Raised;
            FPValue = Power * (1 + Result);
          }
          sprintf (Buffer, "%g", FPValue);
          SkipLength = Cnt + 4 - CharCnt;                                       /* (The extra byte will be added in the for-loop) */
          Done = TRUE;
        }
      }
    if (Done)
    {
      strcat (RemappedString + StrLength, Buffer);
      StrLength += strlen (Buffer);
      RealLength += strlen (Buffer);
    }
    else if (DropTABs && SpectrumString[CharCnt] == 0x06)
    {
      strcat (RemappedString + (StrLength ++), " ");
      RealLength ++;
    }
    else
    {
      strcat (RemappedString + StrLength, RemapCharacterSet[SpectrumString[CharCnt]].Token);
      StrLength += strlen (RemapCharacterSet[SpectrumString[CharCnt]].Token);
      RealLength += strlen (RemapCharacterSet[SpectrumString[CharCnt]].Token);
    }
    if (SpectrumString[CharCnt] == 0x0E)
      CharCnt += 5;
    else
      CharCnt += SkipLength;
  }
  *RawLength = RealLength;
  return (StrLength);
}

void RemapSpectrumUDGs (bool Assume48K)

/**********************************************************************************************************************************/
/* Pre   : `Assume48K' is TRUE if the tape is for a 48K.                                                                          */
/* Post  : The UDG characters have been adjusted.                                                                                 */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  word MapSize;

  SelectFont (SpectrumFont, FONT_NORMAL);                                                /* Ensure the Spectrum font is in memory */
  RemapCharacterSet = TokenMap;
  if (Assume48K)
  {
    MapSize = ('V' -'A') * 8;
    memcpy (SpectrumFont->NormalData + (SpectrumFont->Height * 144) + 8, SpectrumUDGs, MapSize);
    strcpy (TokenMap[163].Token, "\xA3");                                                            /* 'T' and 'U' exist as UDGs */
    TokenMap[163].Flags = 0x00;
    strcpy (TokenMap[164].Token, "\xA4");
    TokenMap[164].Flags = 0x00;
  }
  else
  {
    MapSize = ('T' -'A') * 8;
    memcpy (SpectrumFont->NormalData + (SpectrumFont->Height * 144) + 8, SpectrumUDGs, MapSize);
    strcpy (TokenMap[163].Token, "SPECTRUM");
    TokenMap[163].Flags = F_PAD_L | F_PAD_R;
    strcpy (TokenMap[164].Token, "PLAY");
    TokenMap[164].Flags = F_PAD_L | F_PAD_R;
  }
}
