/**********************************************************************************************************************************/
/* Project      : TAPER                                                                                                           */
/* Module       : TPMAIN.C                                                                                                        */
/* Description  : Main program                                                                                                    */
/* Version type : Main program                                                                                                    */
/* Last changed : 09-12-2001  11:00                                                                                               */
/* Update count : 30                                                                                                              */
/* OS type      : PC (DOS)                                                                                                        */
/* Compiler     : Watcom C32 Optimizing Compiler Version 10.x                                                                     */
/* Linker       : Watcom C Linker                                                                                                 */
/*                                                                                                                                */
/*                Copyleft (C) 1997-2001 ThunderWare Research Center, written by Martijn van der Heide.                           */
/*                                                                                                                                */
/*                This program is free software; you can redistribute it and/or                                                   */
/*                modify it under the terms of the GNU General Public License                                                     */
/*                as published by the Free Software Foundation; either version 2                                                  */
/*                of the License, or (at your option) any later version.                                                          */
/*                                                                                                                                */
/*                This program is distributed in the hope that it will be useful,                                                 */
/*                but WITHOUT ANY WARRANTY; without even the implied warranty of                                                  */
/*                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                                   */
/*                GNU General Public License for more details.                                                                    */
/*                                                                                                                                */
/*                You should have received a copy of the GNU General Public License                                               */
/*                along with this program; if not, write to the Free Software                                                     */
/*                Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                     */
/*                                                                                                                                */
/**********************************************************************************************************************************/

#define __TPMAIN_MAIN__

#include "wbench32/wbench32.h"
#include <dos.h>
#include <fcntl.h>
#include <io.h>
#include <direct.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "tpfile.h"
#include "tpscreen.h"
#include "tpconfig.h"
#include "tpmenu.h"
#include "tpedit.h"
#include "tpaudio.h"
#include "tpserpar.h"
#include "tpdisass.h"
#include "tpmain.h"

#pragma pack                ()

/**********************************************************************************************************************************/
/* Define the spectrum palette for VGA                                                                                            */
/**********************************************************************************************************************************/

dword      SpecPalette[16]  = { 0x00000000, 0x002F0000, 0x00002F00, 0x002F2F00, 0x0000002F, 0x002F002F, 0x00002F2F, 0x002A2A2A,
                                0x001A1A1A, 0x003F0000, 0x00003F00, 0x003F3F00, 0x0000003F, 0x003F003F, 0x00003F3F, 0x003F3F3F };

//short     SpecColTable[16]  = { 0, 1, 4, 5, 2, 3, 6, 7, 8, 9, 12, 13, 10, 11, 14, 15 };

static void ActivateTapeEntry (short WhichWindow);

/**********************************************************************************************************************************/
/* Define the function prototypes                                                                                                 */
/**********************************************************************************************************************************/

int  HandleConfig    (struct WindowInfo *Window, struct ButtonInfo *Button);
int  HandleShell     (struct WindowInfo *Window, struct ButtonInfo *Button);
int  HandleAbout     (struct WindowInfo *Window, struct ButtonInfo *Button);
int  HandleSysInfo   (struct WindowInfo *Window, struct ButtonInfo *Button);
int  HandleHelp      (struct WindowInfo *Window, struct ButtonInfo *Button);

/**********************************************************************************************************************************/
/* Define the global variables                                                                                                    */
/**********************************************************************************************************************************/

char                  CTapeSize[2][10] = {{"\0"},{"\0"}};
struct FontInfo       SpecFont     = {"SPECTRUM", 8, 8};
struct FontInfo       SpecSmall    = {"SPEC6X8", 6, 8};
struct FontInfo       UsedMenuFont = {"8X10", 8, 10};
struct WindowInfo     MainWindow   = {0, 0, 640, 480, 640, 480, 640, 480, 0, WINB_NONE, DBLACK, LBLACK, NULL, WINF_NOBUFFER,
                                      NULL, NULL};
struct WindowInfo     Overlay[2]   = {{10, 27, 273, 440, 273, 440, 273, 440, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE,
                                       NULL, WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_BACKDROP, NULL, "SOURCE TAPE"},
                                      {357, 27, 273, 440, 273, 440, 273, 440, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE,
                                       NULL, WINF_HUSE | WINF_HCENTER | WINF_BACKDROP, NULL, "DESTINATION TAPE"}};
struct ButtonInfo     MTPSize[2]   = {{BT_TEXT, BC_TEXT, BF_FWIDTH | BF_FITTXT, 205, 10, 8, DBLACK, DWHITE, NULL, 0,
                                       CTapeSize[TWLEFT], NULL},
                                      {BT_TEXT, BC_TEXT, BF_FWIDTH | BF_FITTXT, 205, 10, 8, DBLACK, DWHITE, NULL, 0,
                                       CTapeSize[TWRIGHT], NULL}};
struct ButtonInfo     MTPName[2]   = {{BT_TEXT, BC_TEXT, BF_FWIDTH | BF_FITDIR, 0, 428, 43, DBLACK, DWHITE, NULL, 0,
                                       TapeWindow[TWLEFT].ConvTapeName, NULL},
                                      {BT_TEXT, BC_TEXT, BF_FWIDTH | BF_FITDIR, 0, 428, 43, DBLACK, DWHITE, NULL, 0,
                                       TapeWindow[TWRIGHT].ConvTapeName, NULL}};
struct ButtonInfo     MTPChange[2] = {{BT_TEXT, BC_GADGET, BF_FWIDTH, 265, 428, 1, DRED, DWHITE, NULL, 0, "\023", NULL},
                                      {BT_TEXT, BC_GADGET, BF_FWIDTH, 265, 428, 1, DRED, DWHITE, NULL, 0, "\023", NULL}};
struct ButtonInfo     MTP128K[2]   = {{BT_BOOL, BC_GADGET, BF_BCLEAR, 7, 10, 0, DBLACK, DWHITE, NULL, 1,
                                       &TapeWindow[TWLEFT].ModeIs128K, NULL},
                                      {BT_BOOL, BC_GADGET, BF_BCLEAR, 7, 10, 0, DBLACK, DWHITE, NULL, 2,
                                       &TapeWindow[TWRIGHT].ModeIs128K, NULL}};
struct ButtonInfo     MConfig      = {BT_PUSH, BC_TEXT, BF_FWIDTH, 294, 162, 8, DBLACK, DWHITE, NULL, 3, "@Config", HandleConfig};
struct ButtonInfo     MShell       = {BT_PUSH, BC_TEXT, BF_FWIDTH, 294, 180, 8, DBLACK, DWHITE, NULL, 4, "S@hell", HandleShell};
struct ButtonInfo     MQuit        = {BT_PUSH, BC_TEXT, BF_FWIDTH, 294, 198, 8, DBLACK, DWHITE, NULL, 5, "@Quit", NULL};
struct ButtonInfo     MHelp        = {BT_PUSH, BC_TEXT, BF_FWIDTH, 294, 382, 8, DBLACK, DWHITE, NULL, 6, "Help", HandleHelp};
struct ButtonInfo     MAbout       = {BT_PUSH, BC_TEXT, BF_FWIDTH, 294, 400, 8, DBLACK, DWHITE, NULL, 7, "A@bout", HandleAbout};
struct ButtonInfo     MSysInfo     = {BT_PUSH, BC_TEXT, BF_FWIDTH, 294, 418, 8, DBLACK, DWHITE, NULL, 8, "S@ysInfo", HandleSysInfo};

char                  PathBuffer[MAXDIR];
char                  LastAddPath[MAXDIR];
char                  LastAddFile[13];
bool                  LastAddFileSet       = FALSE;

void SurroundScreen (struct WindowInfo *Window, short StartX, short StartY)

/**********************************************************************************************************************************/
/* Pre   : `Window' points to the window where the screen is, `StartX' and `StartY' should both be 2 pixels less than the screen. */
/* Post  : A shadow box has been drawn around the screen.                                                                         */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  short EndX;
  short EndY;

  EndX = StartX + 259;
  EndY = StartY + 195;
  WDrawLine (Window, SystemColors.ShadowDark, StartX, StartY, EndX, StartY);
  WDrawLine (Window, SystemColors.ShadowDark, StartX, StartY, StartX, EndY);
  WDrawLine (Window, SystemColors.ShadowLight, StartX, EndY, EndX, EndY);
  WDrawLine (Window, SystemColors.ShadowLight, EndX, StartY, EndX, EndY);
}

void RescueHandler (int ExitCode, char *ErrorMessage)

/**********************************************************************************************************************************/
/* Pre   : This is the pre-fatal anything-but-low-memory handler.                                                                 */
/* Post  : The database has been rescued if changed and the user accepted.                                                        */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  unsigned Dummy;

  ExitCode = ExitCode; /* keep compiler happy */
  if (TapeWindow[TWLEFT].TapeChanged || TapeWindow[TWRIGHT].TapeChanged)
  {
    SoundErrorBeeper (BEEP_LONG);
    if (!QuestionRequester (-1, -1, FALSE, FALSE, NULL, NULL, NULL, NULL, NULL, "Rescue Tapes",
        "%s\n\nTry to rescue changed tapes in xxxTAPE.SAV ?", ErrorMessage))
    {
      if (TapeWindow[TWLEFT].TapeChanged)
        WriteTape (TWLEFT, TRUE);
      if (TapeWindow[TWRIGHT].TapeChanged)
        WriteTape (TWRIGHT, TRUE);
      _dos_setdrive (CurDrv, &Dummy);                                                       /* Return to home drive and directory */
      chdir (CurPath);
      sprintf (ErrorMessage, "Tapes rescued in xxxTAPE.SAV");
    }
    else
      sprintf (ErrorMessage, "All changes are lost");
  }
}

void LowMemHandler (int MemoryType, char *ErrorMessage)

/**********************************************************************************************************************************/
/* Pre   : This is the pre-fatal low memory handler.                                                                              */
/* Post  : The database has been rescued if changed and the user accepted.                                                        */
/* Import: RescueHandler.                                                                                                         */
/**********************************************************************************************************************************/

{
  MemoryType = MemoryType;
  RescueHandler (1, ErrorMessage);                                                                          /* (Memory error = 1) */
}

static void ActivateTapeEntry (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : An entry in `WhichWindow' has been appointed (with SelectTapeEntry or by clicking)                                     */
/* Post  : The menu has been adjusted to the type of entry.                                                                       */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct ComboEntry ComEntry;

  GetComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &ComEntry, TapeWindow[WhichWindow].SelectedLine);
  if (!((struct TapeEntry_s *)ComEntry.Identifier)->IsIdentifiedData &&
       ((struct TapeEntry_s *)ComEntry.Identifier)->BlockTypeID != 0x14)                                           /* (Pure Data) */
    GhostMenuViewAs (WhichWindow);                                                               /* Cannot view unidentified data */
  else
    UnghostMenuViewAs (WhichWindow);
  if (((struct TapeEntry_s *)ComEntry.Identifier)->BlockTypeID != 0x15)                                     /* (Direct Recording) */
    GhostMenuDecDirRec (WhichWindow);                                                               /* Cannot decode other blocks */
  else
    UnghostMenuDecDirRec (WhichWindow);
  if (!((struct TapeEntry_s *)ComEntry.Identifier)->Flags.IsHeader)
    GhostMenuEdHeader (WhichWindow);
  else
    UnghostMenuEdHeader (WhichWindow);
  ExpandCollapse (WhichWindow, (struct TapeEntry_s *)ComEntry.Identifier);
}

void UpdateTapeSize (short WhichWindow, long NewSize)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `NewSize' is the new size in memory of the tape.                            */
/* Post  : The Size button has been adjusted. If `NewSize' < 0, it means an empty tape and the button is simply cleared.          */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (NewSize < 0)                                                                                                /* Empty tape ? */
    CTapeSize[WhichWindow][0] = '\0';
  else
    sprintf (CTapeSize[WhichWindow], "%8ld", NewSize);
  UpdateButtonText (&(Overlay[WhichWindow]), &(MTPSize[WhichWindow]), BF_FITTXT);
}

bool TestFileOverwrite (char *Path)

/**********************************************************************************************************************************/
/* Pre   : `Path' points to the full path of the file to tested for existence.                                                    */
/* Post  : If the file existed, the user has been asked for confirmation. FALSE if returned if it didn't exist or the user        */
/*         accepted to overwrite.                                                                                                 */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct stat    StatInfo;
  short register Cnt;

  if (!TPConfig.ConfirmOverwrite)                                                                                 /* Don't care ? */
    return (FALSE);
  Cnt = strlen (Path);
  while (-- Cnt > 0 && Path[Cnt] != '\\')                                                                    /* Find the filename */
    ;
  if (!stat (Path, &StatInfo))                                                                                   /* File exists ? */
    return (QuestionRequester (-1, -1, FALSE, FALSE, "@Ok", "@Cancel", NULL, NULL, NULL, "Confirm Overwrite",
            "File %s exists!\nReplace the existing file?", Path + Cnt + 1, StatInfo.st_size));
  else
    return (FALSE);
}

void AutoInsertNewTape (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' holds either TWLEFT or TWRIGHT.                                                                          */
/* Post  : If no tape was in the window yet, an empty one is loaded now.                                                          */
/* Import: SetTapeName.                                                                                                           */
/**********************************************************************************************************************************/

{
  if (TapeWindow[WhichWindow].NoTape)
  {
    TapeWindow[WhichWindow].NoTape = FALSE;
    TapeWindow[WhichWindow].NonexistingTape = TRUE;
    TapeWindow[WhichWindow].NewTape = TRUE;
    TapeWindow[WhichWindow].EmptyTape = TRUE;
    TapeWindow[WhichWindow].TapeFloppyId = 0;
    TapeWindow[WhichWindow].BackupTape = FALSE;
    SetTapeName (WhichWindow, TAPENONE, TRUE);
    UnghostMenu (WhichWindow);
    GhostMenuBackup (WhichWindow);
  }
}

void SelectTapeEntry (short WhichWindow, comboent WhichEntry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `WhichEntry' holds the entry to be selected.                                */
/* Post  : The entry has been selected and all buttons/menu entries (un)ghosted accordingly.                                      */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct ComboEntry ComEntry;

  TapeWindow[WhichWindow].SelectedLine = WhichEntry;                                     /* Make sure we point to the right thing */
  SelectEntry (&(TapeWindow[WhichWindow].IndexWindow), WhichEntry);
  GetComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &ComEntry, WhichEntry);
  if (!((struct TapeEntry_s *)ComEntry.Identifier)->Flags.IsHeader)
    GhostMenuEdHeader (WhichWindow);
  else
    UnghostMenuEdHeader (WhichWindow);
  if (((struct TapeEntry_s *)ComEntry.Identifier)->BlockTypeID != 0x15)                                     /* (Direct Recording) */
    GhostMenuDecDirRec (WhichWindow);                                                               /* Cannot decode other blocks */
  else
    UnghostMenuDecDirRec (WhichWindow);
  if (!((struct TapeEntry_s *)ComEntry.Identifier)->IsIdentifiedData)
    GhostMenuViewAs (WhichWindow);                                                               /* Cannot view unidentified data */
  else
    UnghostMenuViewAs (WhichWindow);
}

void SelectNewTapeEntry (short WhichWindow, comboent NewEntry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `WhichEntry' holds the entry to be selected.                                */
/* Post  : The entry is checked to point to a tape block. If not, it's the first line of a multi-line entry, in which case the    */
/*         entry number is either moved forward until satisfied, or moved one entry back if the end of the tape was reached.      */
/* Import: SelectTapeEntry.                                                                                                       */
/**********************************************************************************************************************************/

{
  struct ComboEntry CEntry;
  comboent          RealEntry;
  bool              Done = FALSE;

  if (NewEntry > TapeWindow[WhichWindow].IndexWindow.NumEntries)                  /* First make sure the entry number is in range */
    NewEntry = TapeWindow[WhichWindow].IndexWindow.NumEntries;
  RealEntry = NewEntry;
  if (RealEntry < 0)                                                                                         /* Tape is now empty */
  {
    TapeWindow[WhichWindow].SelectedLine = -1;
    TapeWindow[WhichWindow].IndexWindow.LastSelected = -1;
    return;
  }
  while (!Done)
  {
    GetComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &CEntry, RealEntry);
    if (!(CEntry.Selected & ENT_NOSELECT) && CEntry.Identifier != 0)                                  /* Points to a data block ? */
    {
      SelectTapeEntry (WhichWindow, RealEntry);
      Done = TRUE;
    }
    else if (RealEntry < TapeWindow[WhichWindow].IndexWindow.NumEntries)                                  /* Nope, try next entry */
      RealEntry ++;
    else                                                                                                   /* End of tape reached */
    {
      if (NewEntry > 0)
        SelectTapeEntry (WhichWindow, NewEntry - 1);
      Done = TRUE;
    }
  }
}

short GetFileType (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The window's variable `FileType' has been set according to the spectrum file. This value is also returned.             */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  char register *PCNameExt;

  PCNameExt = TapeWindow[WhichWindow].TapeName + strlen (TapeWindow[WhichWindow].TapeName) - 4;
  if      (!strcmp (PCNameExt, ".TAP")) TapeWindow[WhichWindow].FileType = FT_TAPZ80;                             /* Tape formats */
  else if (!strcmp (PCNameExt, ".ZXS")) TapeWindow[WhichWindow].FileType = FT_ZXS;
  else if (!strcmp (PCNameExt, ".TZX")) TapeWindow[WhichWindow].FileType = FT_TZX;
  else if (!strcmp (PCNameExt, ".VOC")) TapeWindow[WhichWindow].FileType = FT_VOC;
  else if (!strcmp (PCNameExt, ".LTP")) TapeWindow[WhichWindow].FileType = FT_LTP;
  else if (!strcmp (PCNameExt, ".RAW")) TapeWindow[WhichWindow].FileType = FT_ACORNRAW;
  else if (!strcmp (PCNameExt, ".CSW")) TapeWindow[WhichWindow].FileType = FT_CSW;
  else if (!strcmp (PCNameExt, ".Z80")) TapeWindow[WhichWindow].FileType = FT_Z80;                            /* Snapshot formats */
  else if (!strcmp (PCNameExt, ".SNA")) TapeWindow[WhichWindow].FileType = FT_SNA;
  else if (!strcmp (PCNameExt, ".WAV")) TapeWindow[WhichWindow].FileType = FT_BADSOUND;                     /* Wrong sound format */
  else                                  TapeWindow[WhichWindow].FileType = FT_UNKNOWN;
  return (TapeWindow[WhichWindow].FileType);
}

void SignalTapeMode (short WhichWindow, bool ModeIs128K)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `ModeIs128K' holds the new value.                                           */
/* Post  : The flag has been exported and the '128K' button has been updated.                                                     */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  TapeWindow[WhichWindow].ModeIs128K = ModeIs128K;
  MTP128K[WhichWindow].Contents[0] = ModeIs128K;
  UpdateButtonText (&(Overlay[WhichWindow]), &(MTP128K[WhichWindow]), 0);
}

void SignalTapeChanged (short WhichWindow, bool Changed)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Changed' holds the new value.                                              */
/* Post  : The flag has been exported and the 'changed' button has been updated.                                                  */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  TapeWindow[WhichWindow].TapeChanged = Changed;
  MTPChange[WhichWindow].Contents[0] = Changed ? '\011' : '\023';
  UpdateButtonText (&(Overlay[WhichWindow]), &(MTPChange[WhichWindow]), 0);
}

void GhostEmptyTape (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT and the tape in that window is empty (or no tape at all).                    */
/* Post  : All appropriate buttons and menu entries have been ghosted.                                                            */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  GhostMenuEmptyTape (WhichWindow);                                                             /* Ghost appropriate menu entries */
}

void UnghostTape (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT and the tape in that window is no longer empty.                              */
/* Post  : All appropriate buttons and menu entries have been unghosted.                                                          */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  UnghostMenu (WhichWindow);                                                                         /* Activate the menu entries */
}

void SetTapeName (short WhichWindow, char *FileName, bool UpdateButton)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Pre   : `FileName' holds the (new) name for the tape to be handled.                                                            */
/*         `UpdateButton' is TRUE if the button to which the name is linked should be updated as well.                            */
/* Post  : The `TapeName', `TapeBackupName' and `TapeTempName' variables have been set to this name.                              */
/* Import: GetFileType.                                                                                                           */
/**********************************************************************************************************************************/

{
  char Drive[4];
  char Path[MAXDIR];
  char FName[9];
  char Ext[5];

  if (TapeWindow[WhichWindow].NewTape)
  {
    strcpy (TapeWindow[WhichWindow].TapeName, FileName);
    strcpy (TapeWindow[WhichWindow].ConvTapeName, FileName);
  }
  else
  {
    if (_fullpath (TapeWindow[WhichWindow].TapeName, FileName, MAXDIR) == NULL)
      FatalError (7, "Invalid path %s", FileName);
    ConvertPath (TapeWindow[WhichWindow].TapeName, TapeWindow[WhichWindow].ConvTapeName, TapeWindow[WhichWindow].TapeFloppyId);
  }
  _splitpath (TapeWindow[WhichWindow].TapeName, Drive, Path, FName, Ext);
  _makepath (TapeWindow[WhichWindow].TempName, Drive, Path, FName, TAPETEMP);
  _makepath (TapeWindow[WhichWindow].BackupName, Drive, Path, FName, TAPEBACKUP);
  GetFileType (WhichWindow);
  if (UpdateButton)
    UpdateButtonText (&(Overlay[WhichWindow]), &(MTPName[WhichWindow]), BF_FITDIR);
}

void DeselectAll (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : All tape blocks in that window have been deselected.                                                                   */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct ComboEntry Entry;
  comboent register Cnt;

  if (!TapeWindow[WhichWindow].EmptyTape)
    for (Cnt = 0 ; Cnt <= TapeWindow[WhichWindow].IndexWindow.NumEntries ; Cnt ++)
    {
      GetComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &Entry, Cnt);
      Entry.Selected &= ~SELECTBIT;
      if (Entry.Identifier != 0)                                                                 /* It's a tape directory entry ? */
        SetTapeBlockSelectFlag ((struct TapeEntry_s *)Entry.Identifier, 0);                                        /* Then update */
      PutComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &Entry, Cnt, TRUE);
    }
}

int HandleConfig (struct WindowInfo *Window, struct ButtonInfo *Button)

/**********************************************************************************************************************************/
/* Pre   : The `Config' button has been pressed (or released).                                                                    */
/* Post  : The configuration has been changed.                                                                                    */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  bool OldRemapBrightBlack;

  Window = Window; /* keep compiler happy */
  if (!(Button->Flags & BF_PRESSED))
  {
    OldRemapBrightBlack = TPConfig.RemapBrightBlack;                                            /* Store some important variables */
    ChangeConfig ();                                                                /* Allow to make changes to the configuration */
    if (ActiveWindow == &(Overlay[TWLEFT]))                                                          /* See if we can play things */
      GhostMenuNoSoundBlaster (TWLEFT); 
    else
      GhostMenuNoSoundBlaster (TWRIGHT); 
    if (SystemDefaults.SyncSlider)                                                                           /* Sync slider bar ? */
    {
      TapeWindow[TWLEFT].IndexWindow.Flags |= COMF_SYNCSLIDER;
      TapeWindow[TWRIGHT].IndexWindow.Flags |= COMF_SYNCSLIDER;
    }
    else
    {
      TapeWindow[TWLEFT].IndexWindow.Flags &= ~COMF_SYNCSLIDER;
      TapeWindow[TWRIGHT].IndexWindow.Flags &= ~COMF_SYNCSLIDER;
    }
  }
  return (WINH_NONE);
}

int HandleShell (struct WindowInfo *Window, struct ButtonInfo *Button)

/**********************************************************************************************************************************/
/* Pre   : The `Shell' button has been pressed (or released).                                                                     */
/* Post  : A shell has been started and shut down by the user.                                                                    */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  unsigned Dummy;

  Window = Window; /* keep compiler happy */
  if (!(Button->Flags & BF_PRESSED))
  {
    SwapOut (TRUE, "", "", NULL, "Enter 'EXIT' to return to Taper\n\n", FALSE, FALSE, FALSE);                        /* Shell out */
    _dos_setdrive (CurDrv, &Dummy);                                                                     /* Restore drive and path */
    chdir (CurPath);
  }
  return (WINH_NONE);
}

int HandleAbout (struct WindowInfo *Window, struct ButtonInfo *Button)

/**********************************************************************************************************************************/
/* Pre   : The `About' button has been pressed (or released).                                                                     */
/* Post  : An about box has been shown to the user.                                                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  Window = Window; /* keep compiler happy */
  if (!(Button->Flags & BF_PRESSED))
    AboutBox (TRUE);
  return (WINH_NONE);
}

int HandleHelp (struct WindowInfo *Window, struct ButtonInfo *Button)

/**********************************************************************************************************************************/
/* Pre   : `Config.EditPath' is a non-empty string.                                                                               */
/* Post  : The `Config.EditPath' has been started on SGD.DOC (elementary help construction)                                       */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  char DocPath[MAXDIR];

  Window = Window; /* keep compiler happy */
  if (!(Button->Flags & BF_PRESSED))
  {
    if (!RequestFloppy (ProgramDrive[0], ResourceFloppyId))
      return (WINH_NONE);
    sprintf (DocPath, "%s%sreadme", ProgramDrive, ProgramPath);
    FileViewRequester (80, 100, 128, 80, 80, 25, DocPath, TRUE, "Taper Help");
  }
  return (WINH_NONE);
}

int HandleSysInfo (struct WindowInfo *Window, struct ButtonInfo *Button)

/**********************************************************************************************************************************/
/* Pre   : The `SysInfo' button has been pressed (or released).                                                                   */
/* Post  : A system info box has been shown to the user.                                                                          */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  Window = Window; /* keep compiler happy */
  if (!(Button->Flags & BF_PRESSED))
    SysInfoBox ();
  return (WINH_NONE);
}

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> MAIN PROGRAM <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

void main (int argc, char **argv)
{
  struct ComboEntry   ComEntry;
  struct ButtonInfo  *ButtonP;
  struct TapeEntry_s *DirP;
  comboent register   Cnt;
  short               WhichTape          = TWLEFT;
  bool                RequestQuit        = FALSE;
  bool                TimeToQuit         = FALSE;
  bool                CancelQuit         = TRUE;
  bool                Found;
  bool                DisableMSWin       = FALSE;
  bool                HaveSoundcard      = TRUE;
  bool                WriteThisTape;
  int                 ArgCnt             = 0;
  word                BlockSize;
  bool                MoreBlocks;
  dword               BlockIndex;
  dword               Dummy;
  comboent            BlockNo;
  comboent            NewSelectedEntry;

  __minreal = 100*1024;                                                                 /* Needed to spawn other DOS/4GW programs */

  UseDecryptUserValues = FALSE;
  DecryptXOR = 0x98;                                                                                      /* (SpeedLock 2 values) */
  DecryptADD = 0x0B;
  InitVOCWriteVars = TRUE;
  _MustInitializeJumpCodes = TRUE;
  memset (&(TapeWindow[TWLEFT]), 0, sizeof (struct TapeWindow_s));
  memset (&(TapeWindow[TWRIGHT]), 0, sizeof (struct TapeWindow_s));
  TapeWindow[TWLEFT].AutoDecryptBlocks = TRUE;
  TapeWindow[TWRIGHT].AutoDecryptBlocks = TRUE;
  TapeWindow[TWLEFT].DecryptAnything = FALSE;
  TapeWindow[TWRIGHT].DecryptAnything = FALSE;
  TapeWindow[TWLEFT].NoTape = TRUE;
  TapeWindow[TWRIGHT].NoTape = TRUE;
  TapeWindow[TWLEFT].FileType = FT_UNKNOWN;
  TapeWindow[TWRIGHT].FileType = FT_UNKNOWN;
  TapeWindow[TWLEFT].NonexistingTape = TRUE;
  TapeWindow[TWRIGHT].NonexistingTape = TRUE;
  TapeWindow[TWLEFT].NumEntries = -1;
  TapeWindow[TWRIGHT].NumEntries = -1;
  TapeWindow[TWLEFT].Tape.XMSHandle = -1;
  TapeWindow[TWRIGHT].Tape.XMSHandle = -1;
  ReleaseTape (TWLEFT);
  ReleaseTape (TWRIGHT);
  _HaveMixer = TRUE;
  while (++ ArgCnt < argc)                                                                       /* Handle command line arguments */
    if (argv[ArgCnt][0] == '-' || argv[ArgCnt][0] == '/')                                                           /* A switch ? */
      switch (toupper (argv[ArgCnt][1]))
      {
        case 'W' : DisableMSWin = TRUE; break;
        case 'U' : break;                                                                                           /* (obsolete) */
        case 'S' : HaveSoundcard = FALSE; break;
        case 'M' : _HaveMixer = FALSE; break;
        case '?' :
        case 'H' : fprintf (stderr, "SYNTAX: TAPER [-W] [-S] [-M] [<tapefile>] [<tapefile>]\n");
                   fprintf (stderr, "        -W : disable Windows \`multitasking\'\n");
                   fprintf (stderr, "        -S : do not check for a soundcard\n");
                   fprintf (stderr, "        -M : do not check for a software mixer on the soundcard\n");
                   fprintf (stderr, "See README and COPYING for details\n");
                   exit (0);
        default  : fprintf (stderr, "FATAL ERROR - Unknown switch \'%s\'\n", argv[ArgCnt] + 1);
                   exit (1);
      }
    else                                                                                        /* A database name has been given */
    {
      SetTapeName (WhichTape, strupr (argv[ArgCnt]), FALSE);
      if (TapeWindow[WhichTape].FileType == FT_UNKNOWN)
      {
        fprintf (stderr, "Error: %s is not a Spectrum tape file\n", argv[ArgCnt]);
        exit (1);
      }
      if (TapeWindow[WhichTape].FileType == FT_BADSOUND)
      {
        fprintf (stderr, "Error: %s is not not understood.\n"
                 "       Please convert to (8-bit mono) .VOC format\n", argv[ArgCnt]);
        exit (1);
      }
      TapeWindow[WhichTape].NonexistingTape = FALSE;                                         /* Assume a valid filename was given */
      TapeWindow[WhichTape].NoTape = FALSE;
      WhichTape = (WhichTape == TWLEFT ? TWRIGHT : TWLEFT);
    }
  InitGUI (argv[0], LBLACK, SpecPalette, DisableMSWin, HaveSoundcard, _HaveMixer);
  AllocateScreenBuffer ();
  AllocateAudioBuffer ();
  _makepath (ConfigFile, ProgramDrive, ProgramPath, ProgramName, CONFIGEXT);              /* Build name of the configuration file */
  ReadConfig ();                                                                                        /* Read the configuration */
  if (TPConfig.DirectVideoControl)                                                         /* The RIL must not be used if found ? */
    Screen.UseRIL = FALSE;
  if (TPConfig.UseCustomPalette)                                                            /* User want his own custom palette ? */
  {
    SpecPalette[0]  = TPConfig.CustomColor00;   /* VGA  0: DBLACK   */
    SpecPalette[1]  = TPConfig.CustomColor01;   /* VGA  1: DBLUE    */
    SpecPalette[2]  = TPConfig.CustomColor04;   /* VGA  2: DGREEN   */
    SpecPalette[3]  = TPConfig.CustomColor05;   /* VGA  3: DCYAN    */
    SpecPalette[4]  = TPConfig.CustomColor02;   /* VGA  4: DRED     */
    SpecPalette[5]  = TPConfig.CustomColor03;   /* VGA  5: DMAGENTA */
    SpecPalette[6]  = TPConfig.CustomColor06;   /* VGA  6: DYELLOW  */
    SpecPalette[7]  = TPConfig.CustomColor07;   /* VGA  7: DWHITE   */
    SpecPalette[8]  = TPConfig.CustomColor08;   /* VGA  8: LBLACK   */
    SpecPalette[9]  = TPConfig.CustomColor09;   /* VGA  9: LBLUE    */
    SpecPalette[10] = TPConfig.CustomColor12;   /* VGA 10: LGREEN   */
    SpecPalette[11] = TPConfig.CustomColor13;   /* VGA 11: LCYAN    */
    SpecPalette[12] = TPConfig.CustomColor10;   /* VGA 12: LRED     */
    SpecPalette[13] = TPConfig.CustomColor11;   /* VGA 13: LMAGENTA */
    SpecPalette[14] = TPConfig.CustomColor14;   /* VGA 14: LYELLOW  */
    SpecPalette[15] = TPConfig.CustomColor15;   /* VGA 15: LWHITE   */
    RemapAllPalette (SpecPalette);                                                                 /* Activate the custom palette */
  }
  if (TPConfig.BlankerOn)
    InstallBlanker (TPConfig.BlankerTimeOut, NULL);
  else
    DisableBlanker ();
  strcpy (LastAddPath, CurPath);
  LastAddFile[0] = '\0';
  _UsedMenuFont = &UsedMenuFont;
  SpectrumFont = &SpecFont;                                                      /* Export the Spectrum font to the other modules */
  SpectrumSmallFont = &SpecSmall;
  if (TPConfig.PreLoadResources)
  {
    SelectFont (&SpecFont, FONT_NORMAL);
    SelectFont (&SpecSmall, FONT_NORMAL);
    SelectFont (DefaultFont, FONT_NORMAL);
    SelectFont (DefaultFont, FONT_BOLD);
    SelectFont (DefaultFont, FONT_ITALICS);
    SelectFont (_UsedMenuFont, FONT_NORMAL);
  }
  LoadHardwareTable ();                                                                       /* Load the HARDWARE.INC table file */

  if (!TapeWindow[TWLEFT].NonexistingTape || !TapeWindow[TWRIGHT].NonexistingTape)  /* A filename was given on the command line ? */
  {
    AboutBox (FALSE);
    _PreReadingTape = TRUE;                                                                   /* Signal: no windows to update yet */
    if (!TapeWindow[TWLEFT].NonexistingTape)
      if (!ReadTape (TWLEFT, FALSE))                                                                /* Read the database into XMS */
        FatalError (10, "Tape file corrupt");
      else
      {
        TapeWindow[TWLEFT].NonexistingTape = (TapeWindow[TWLEFT].NumEntries == -1 ? TRUE : FALSE);    /* New (nonexisting) tape ? */
        if (_TAPNoise)
        {
          _TAPNoise = FALSE;
          TapeWindow[TWLEFT].TapeChanged = TRUE;
          MTPChange[TWLEFT].Contents[0] = '\011';
          ErrorRequester (-1, -1, NULL, NULL, "Skipped Noise", "Skipped noise blocks in source tape");
        }
      }
    if (!TapeWindow[TWRIGHT].NonexistingTape)
      if (!ReadTape (TWRIGHT, FALSE))                                                               /* Read the database into XMS */
        FatalError (10, "Tape file corrupt");
      else
      {
        TapeWindow[TWRIGHT].NonexistingTape = (TapeWindow[TWRIGHT].NumEntries == -1 ? TRUE : FALSE);
        if (_TAPNoise)
        {
          _TAPNoise = FALSE;
          TapeWindow[TWRIGHT].TapeChanged = TRUE;
          MTPChange[TWRIGHT].Contents[0] = '\011';
          ErrorRequester (-1, -1, NULL, NULL, "Skipped Noise", "Skipped noise blocks in destination tape");
        }
      }
    DestroyWindow (&AboutWindow);
  }
  _PreReadingTape = FALSE;

  SelectedAPen[1] = 4;                                                                                             /* (PC order!) */
  SelectedBPen[2] = 3;                                                                    /* Selected blocks with blue background */
  SelectedAPen[3] = 4;
  SelectedBPen[3] = 3;
  RemapSpectrumUDGs (TRUE);                                                                                /* Initialize the UDGs */

  AddButton (&(Overlay[TWLEFT]), &(MTPSize[TWLEFT]), FALSE);
  AddButton (&(Overlay[TWRIGHT]), &(MTPSize[TWRIGHT]), FALSE);
  AddButton (&(Overlay[TWLEFT]), &(MTPName[TWLEFT]), FALSE);
  AddButton (&(Overlay[TWRIGHT]), &(MTPName[TWRIGHT]), FALSE);
  AddButton (&(Overlay[TWLEFT]), &(MTPChange[TWLEFT]), FALSE);
  AddButton (&(Overlay[TWRIGHT]), &(MTPChange[TWRIGHT]), FALSE);
  AddButton (&(Overlay[TWLEFT]), &(MTP128K[TWLEFT]), FALSE);
  AddButton (&(Overlay[TWRIGHT]), &(MTP128K[TWRIGHT]), FALSE);
  AddButton (&MainWindow, &MConfig, FALSE);
  AddButton (&MainWindow, &MShell, FALSE);
  AddButton (&MainWindow, &MQuit, FALSE);
  AddButton (&MainWindow, &MHelp, FALSE);
  AddButton (&MainWindow, &MAbout, FALSE);
  AddButton (&MainWindow, &MSysInfo, FALSE);
  DrawWindow (&MainWindow, NULL, 0, TRUE);
  DrawWindow (&(Overlay[TWLEFT]), NULL, 0, TRUE);
  DrawWindow (&(Overlay[TWRIGHT]), NULL, 0, TRUE);
  Wprintf (&(Overlay[TWLEFT]), 1, 4, "{+1128K}N");
  Wprintf (&(Overlay[TWRIGHT]), 1, 4, "{+1128K}N");
  Wprintf (&(Overlay[TWLEFT]), 1, 28, "{+1Size}N");
  Wprintf (&(Overlay[TWRIGHT]), 1, 28, "{+1Size}N");
  MakeComboWindow (&(TapeWindow[TWLEFT].IndexWindow), &(Overlay[TWLEFT]), COMT_SELECT, COMF_NOBUFFER | COMF_VSLIDER |
                   COMF_KEYSELECT, Overlay[TWLEFT].StartX, Overlay[TWLEFT].StartY + 33, 0, 0, 81, 43, 38, &SpecSmall, FALSE,
                   0, NULL, DBLACK, DWHITE, SELECTBIT, 0);
  MakeComboWindow (&(TapeWindow[TWRIGHT].IndexWindow), &(Overlay[TWRIGHT]), COMT_SELECT, COMF_NOBUFFER | COMF_VSLIDER |
                   COMF_KEYSELECT, Overlay[TWRIGHT].StartX, Overlay[TWRIGHT].StartY + 33, 0, 0, 81, 43, 38, &SpecSmall, FALSE,
                   0, NULL, DBLACK, DWHITE, SELECTBIT, 0);
  TapeWindow[TWLEFT].IndexWindow.BlinkRate = TPConfig.EmulateFlash ? (TPConfig.SlowFlash ? 960 : 320) : 0;
  TapeWindow[TWRIGHT].IndexWindow.BlinkRate = TPConfig.EmulateFlash ? (TPConfig.SlowFlash ? 960 : 320) : 0;
  LinkComboWindows (&(TapeWindow[TWLEFT].IndexWindow), &(TapeWindow[TWRIGHT].IndexWindow));                  /* Allow drag & drop */
  InstallMenu (&MainWindow);
  SetActiveWindowMenu (&MainWindow, TWLEFT);

  GhostMenuNoSoundBlaster (TWLEFT);
  GhostMenuNoSoundBlaster (TWRIGHT);
  ParallelRunning = FALSE;

  if (TapeWindow[TWLEFT].NonexistingTape)                                              /* No filename given on the command line ? */
  {
    GhostMenuBackup (TWLEFT);                                                                   /* Then there's nothing to backup */
    GhostMenuEmptyTape (TWLEFT);
    TapeWindow[TWLEFT].BackupTape = FALSE;
    PrintComboEntries (&(TapeWindow[TWLEFT].IndexWindow));
  }
  else                                                                      /* Try reading the provided name into the left window */
  {
    SetupNewData (TWLEFT);
    if (TapeWindow[TWLEFT].IndexWindow.NumEntries < 0)                                                            /* Empty tape ? */
      GhostEmptyTape (TWLEFT);                                                                        /* Ghost all action buttons */
    else
      SelectNewTapeEntry (TWLEFT, 0);
  }
  if (TapeWindow[TWRIGHT].NonexistingTape)
  {
    GhostMenuBackup (TWRIGHT);
    GhostMenuEmptyTape (TWRIGHT);
    TapeWindow[TWRIGHT].BackupTape = FALSE;
    PrintComboEntries (&(TapeWindow[TWRIGHT].IndexWindow));
  }
  else
  {
    SetupNewData (TWRIGHT);
    if (TapeWindow[TWRIGHT].IndexWindow.NumEntries < 0)
      GhostEmptyTape (TWRIGHT);
    else
      SelectNewTapeEntry (TWRIGHT, 0);
  }

  if (SystemDefaults.SyncSlider)                                                                             /* Sync slider bar ? */
  {
    TapeWindow[TWLEFT].IndexWindow.Flags |= COMF_SYNCSLIDER;
    TapeWindow[TWRIGHT].IndexWindow.Flags |= COMF_SYNCSLIDER;
  }
  RegisterLowMemHandler (LowMemHandler);                                                       /* Set a low-memory pre-fatal trap */
  RegisterRescueHandler (RescueHandler);                                                       /* Trap all other pre-fatal errors */

  if (_InvalidPort)                                                        /* Bad LPT or COM port selected in the configuration ? */
    ErrorRequester (-1, -1, NULL, NULL, "Bad Port Warning",
                    "{BWARNING}B: You selected a non-existing LPT and/or COM port in your TAPER.INI file.\n"
                    "Please use the \'Config\' button to select an existing port.");

  while (CancelQuit)
  {
    while (!TimeToQuit)                                                                                              /* MAIN LOOP */
    {
      if (!MenuStripActive && Mouse.Left && !Mouse.LeftStillPressed)                                        /* Left mouse click ? */
        if (Mouse.Xx >= Overlay[TWLEFT].A_StartX && Mouse.Xx <= Overlay[TWLEFT].A_EndX &&
            Mouse.Yy >= Overlay[TWLEFT].A_StartY && Mouse.Yy <= Overlay[TWLEFT].A_EndY)
        {
          ActivateWindow (&(Overlay[TWLEFT]));
          SetActiveWindowMenu (&MainWindow, TWLEFT);
        }
        else if (Mouse.Xx >= Overlay[TWRIGHT].A_StartX && Mouse.Xx <= Overlay[TWRIGHT].A_EndX &&
                 Mouse.Yy >= Overlay[TWRIGHT].A_StartY && Mouse.Yy <= Overlay[TWRIGHT].A_EndY)
        {
          ActivateWindow (&(Overlay[TWRIGHT]));
          SetActiveWindowMenu (&MainWindow, TWRIGHT);
        }
      if (GlobalKey == K_TAB)                                                                                  /* [TAB] pressed ? */
      {
        if (ActiveWindow == &(Overlay[TWLEFT]))                                                           /* Switch active window */
        {
          ActivateWindow (&(Overlay[TWRIGHT]));
          SetActiveWindowMenu (&MainWindow, TWRIGHT);
        }
        else if (ActiveWindow == &(Overlay[TWRIGHT]))
        {
          ActivateWindow (&(Overlay[TWLEFT]));
          SetActiveWindowMenu (&MainWindow, TWLEFT);
        }
        GlobalKey = 0x0000;                                                                                    /* Consume the key */
      }
      else if (GlobalKey == K_DEL && !(GlobalShiftStatus & (KS_AALT | KS_ACTRL)))                              /* [DEL] pressed ? */
      {
        DeleteEntryByHand ();
        GlobalKey = 0x0000;
      }
      else if ((GlobalKey & 0x00FF) == '-' && !(GlobalShiftStatus & (KS_AALT | KS_ACTRL)))                       /* [-] pressed ? */
      {
        GlobalKey = 0x0000;
        if (ActiveWindow == &(Overlay[TWLEFT]))
        {
          GetComboEntry (&(TapeWindow[TWLEFT].IndexWindow), &ComEntry, TapeWindow[TWLEFT].SelectedLine);
          if (((struct TapeEntry_s *)ComEntry.Identifier)->BlockTypeID != 0x21 ||       /* (Only applicable to Group/Loop starts) */
              ((struct TapeEntry_s *)ComEntry.Identifier)->BlockTypeID != 0x24 &&
              !(((struct TapeEntry_s *)ComEntry.Identifier)->ComboSelected & EXPANDBIT))
          {
            ((struct TapeEntry_s *)ComEntry.Identifier)->ComboSelected |= EXPANDBIT;                        /* Collapse the block */
            SetupNewData (TWLEFT);                                                                  /* Rebuild the tape directory */
          }
        }
        else if (ActiveWindow == &(Overlay[TWRIGHT]))
        {
          GetComboEntry (&(TapeWindow[TWRIGHT].IndexWindow), &ComEntry, TapeWindow[TWRIGHT].SelectedLine);
          if (((struct TapeEntry_s *)ComEntry.Identifier)->BlockTypeID != 0x21 ||
              ((struct TapeEntry_s *)ComEntry.Identifier)->BlockTypeID != 0x24 &&
              !(((struct TapeEntry_s *)ComEntry.Identifier)->ComboSelected & EXPANDBIT))
          {
            ((struct TapeEntry_s *)ComEntry.Identifier)->ComboSelected |= EXPANDBIT;
            SetupNewData (TWRIGHT);
          }
        }
      }
      else if ((GlobalKey & 0x00FF) == '+' && !(GlobalShiftStatus & (KS_AALT | KS_ACTRL)))                       /* [+] pressed ? */
      {
        GlobalKey = 0x0000;
        if (ActiveWindow == &(Overlay[TWLEFT]))
        {
          GetComboEntry (&(TapeWindow[TWLEFT].IndexWindow), &ComEntry, TapeWindow[TWLEFT].SelectedLine);
          if (((struct TapeEntry_s *)ComEntry.Identifier)->BlockTypeID != 0x21 ||       /* (Only applicable to Group/Loop starts) */
              ((struct TapeEntry_s *)ComEntry.Identifier)->BlockTypeID != 0x24 &&
              (((struct TapeEntry_s *)ComEntry.Identifier)->ComboSelected & EXPANDBIT))
          {
            ((struct TapeEntry_s *)ComEntry.Identifier)->ComboSelected &= ~EXPANDBIT;                         /* Expand the block */
            SetupNewData (TWLEFT);                                                                  /* Rebuild the tape directory */
          }
        }
        else if (ActiveWindow == &(Overlay[TWRIGHT]))
        {
          GetComboEntry (&(TapeWindow[TWRIGHT].IndexWindow), &ComEntry, TapeWindow[TWRIGHT].SelectedLine);
          if (((struct TapeEntry_s *)ComEntry.Identifier)->BlockTypeID != 0x21 ||
              ((struct TapeEntry_s *)ComEntry.Identifier)->BlockTypeID != 0x24 &&
              (((struct TapeEntry_s *)ComEntry.Identifier)->ComboSelected & EXPANDBIT))
          {
            ((struct TapeEntry_s *)ComEntry.Identifier)->ComboSelected &= ~EXPANDBIT;
            SetupNewData (TWRIGHT);
          }
        }
      }
      switch (HandleComboWindow (&(TapeWindow[TWLEFT].IndexWindow), &(TapeWindow[TWLEFT].SelectedLine)))
      {
        case WINH_LINEFIRST   : ActivateTapeEntry (TWLEFT); break;
        case WINH_LINESEL     : ViewTapeBlockInfo (TWLEFT); break;
        case WINH_SELTOGGLE   : GetComboEntry (&(TapeWindow[TWLEFT].IndexWindow), &ComEntry, TapeWindow[TWLEFT].SelectedLine);
                                SetTapeBlockSelectFlag ((struct TapeEntry_s *)ComEntry.Identifier, ComEntry.Selected);
                                TestForSelectedEntries (TWLEFT);
                                break;
        case WINH_MSELTOGGLE  : for (Cnt = 0 ; Cnt <= TapeWindow[TWLEFT].IndexWindow.NumEntries ; Cnt ++)
                                {
                                  GetComboEntry (&(TapeWindow[TWLEFT].IndexWindow), &ComEntry, Cnt);
                                  if (ComEntry.Identifier != 0)
                                    SetTapeBlockSelectFlag ((struct TapeEntry_s *)ComEntry.Identifier, ComEntry.Selected);
                                }
                                TestForSelectedEntries (TWLEFT);
                                break;
        case WINH_DRAGDROPPED : WhichTape = DragTapeBlock ();
                                DeselectAll (TWLEFT);
                                break;
      }
      switch (HandleComboWindow (&(TapeWindow[TWRIGHT].IndexWindow), &(TapeWindow[TWRIGHT].SelectedLine)))
      {
        case WINH_LINEFIRST   : ActivateTapeEntry (TWRIGHT); break;
        case WINH_LINESEL     : ViewTapeBlockInfo (TWRIGHT); break;
        case WINH_SELTOGGLE   : GetComboEntry (&(TapeWindow[TWRIGHT].IndexWindow), &ComEntry, TapeWindow[TWRIGHT].SelectedLine);
                                SetTapeBlockSelectFlag ((struct TapeEntry_s *)ComEntry.Identifier, ComEntry.Selected);
                                TestForSelectedEntries (TWRIGHT);
                                break;
        case WINH_MSELTOGGLE  : for (Cnt = 0 ; Cnt <= TapeWindow[TWRIGHT].IndexWindow.NumEntries ; Cnt ++)
                                {
                                  GetComboEntry (&(TapeWindow[TWRIGHT].IndexWindow), &ComEntry, Cnt);
                                  if (ComEntry.Identifier != 0)
                                    SetTapeBlockSelectFlag ((struct TapeEntry_s *)ComEntry.Identifier, ComEntry.Selected);
                                }
                                TestForSelectedEntries (TWRIGHT);
                                break;
        case WINH_DRAGDROPPED : WhichTape = DragTapeBlock ();
                                DeselectAll (TWRIGHT);
                                break;
      }
      if (GlobalKey == K_CRDN && (GlobalShiftStatus & KS_ACTRL))                                         /* <CTRL><CURSOR DOWN> ? */
      {
        WhichTape = (short)(ActiveWindow == &(Overlay[TWLEFT]) ? TWLEFT : TWRIGHT);                    /* Determine active window */
        GetComboEntry (&(TapeWindow[WhichTape].IndexWindow), &ComEntry, TapeWindow[WhichTape].SelectedLine);
        BlockNo = GetBlockNumber (ComEntry.Text);                                                     /* Get current block number */
        if (ComEntry.Identifier > 0)                                                                       /* Must be a good line */
        {
          DirP = (struct TapeEntry_s *)ComEntry.Identifier;
          if ((DirP->BlockTypeID == 0x21 || DirP->BlockTypeID == 0x24) && (DirP->ComboSelected & EXPANDBIT))
          {                                                                                    /* Entry is collapsed Group/Loop ? */
            DirP->ComboSelected &= ~EXPANDBIT;                                                                /* Expand the block */
            SetupNewData (WhichTape);
          }
          GetBlockTypeInfo (DirP->BlockTypeID, &Found, NULL, NULL, NULL, NULL);
          if (!Found)                                                                                   /* (Only playable blocks) */
          {
            if (++ TapeWindow[WhichTape].SelectedLine > TapeWindow[WhichTape].NumEntries)
              TapeWindow[WhichTape].SelectedLine = 0;
            SelectEntry (&(TapeWindow[WhichTape].IndexWindow), TapeWindow[WhichTape].SelectedLine);
          }
          else
          {
            BlockIndex = GetTapeBlockXMSIndex (WhichTape, DirP);
            SimulatePlayBlock = TRUE;                                             /* Signal: don't actually play it, just advance */
            WriteVOCBlock (WhichTape, -1, DirP, TapeWindow[WhichTape].Tape.XMSHandle, BlockIndex, &Dummy, &BlockNo);
            SimulatePlayBlock = FALSE;
            NewSelectedEntry = FindBlockNumber (WhichTape, BlockNo);
            if (NewSelectedEntry < 0 || NewSelectedEntry > TapeWindow[WhichTape].IndexWindow.NumEntries)
              NewSelectedEntry = 0;
            SelectEntry (&(TapeWindow[WhichTape].IndexWindow), NewSelectedEntry);
            TapeWindow[WhichTape].SelectedLine = NewSelectedEntry;
          }
        }
        GlobalKey = 0x0000;
      }
      if (GlobalKey == K_CRUP && (GlobalShiftStatus & KS_ACTRL))                                           /* <CTRL><CURSOR UP> ? */
      {
        InitVOCWriteVars = TRUE;                                                       /* Signal: clear stacks, loop counters etc */
        SoundErrorBeeper (BEEP_SHORT);                                                     /* Sound the beeper to inform the user */
        GlobalKey = 0x0000;
      }
      HandleWindow (&Overlay[TWLEFT], &ButtonP);                                                    /* Handle the overlay windows */
      HandleWindow (&Overlay[TWRIGHT], &ButtonP);
      switch (HandleWindow (&MainWindow, &ButtonP))                                      /* Handle the MainScreen and all buttons */
      {
        case WINH_MENUNOHAND: switch (((struct MenuEntry *)ButtonP)->ReactNum & 0xFFF8)
                              {
                                case 0xF8 : RequestQuit = TRUE; break;
                                case 0xF0 : WhichTape = (((struct MenuEntry *)ButtonP)->ReactNum & 0x01);
                                            if (InsertTapeEntry (WhichTape))                                          /* 'Insert' */
                                            {
                                              SignalTapeChanged (WhichTape, TRUE);
                                              UnghostMenu (WhichTape);
                                              SetupNewData (WhichTape);                             /* Reprint the tape directory */
                                            }
                                            break;
                                case 0xC0 : WhichTape = (((struct MenuEntry *)ButtonP)->ReactNum & 0x01);
                                            if (EditAllTiming (WhichTape))
                                            {
                                              SignalTapeChanged (WhichTape, TRUE);
                                              UnghostMenu (WhichTape);
                                              SetupNewData (WhichTape);                             /* Reprint the tape directory */
                                            }
                                            break;
                                case 0xA0 : WhichTape = (((struct MenuEntry *)ButtonP)->ReactNum & 0x01);
                                            GetComboEntry (&(TapeWindow[WhichTape].IndexWindow), &ComEntry,
                                                           TapeWindow[WhichTape].SelectedLine);
                                            if (EditHeaderBlock (WhichTape, (struct TapeEntry_s *)ComEntry.Identifier))
                                            {
                                              SignalTapeChanged (WhichTape, TRUE);
                                              UnghostMenu (WhichTape);
                                              SetupNewData (WhichTape);                             /* Reprint the tape directory */
                                            }
                                            break;
                                case 0x08 : ViewTapeBlockInfo (((struct MenuEntry *)ButtonP)->ReactNum & 0x01);
                                            break;
                                //case 0x10 : BlockIndex = 0;
                                            //BlockSize = GetTapeBlock (((struct MenuEntry *)ButtonP)->ReactNum & 0x01,
                                                                      //&BlockIndex, &MoreBlocks);
                                            //ViewSpectrumDump (((struct MenuEntry *)ButtonP)->ReactNum & 0x01, BlockSize);
                                            //break;
                                case 0x10 : RunDisassemblerOrDump ((struct MenuEntry *)ButtonP, FALSE);
                                            break;
                                case 0x20 : BlockIndex = 0;
                                            BlockSize = GetTapeBlock (((struct MenuEntry *)ButtonP)->ReactNum & 0x01,
                                                                      &BlockIndex, &MoreBlocks);
                                            ViewSpectrumHeader (((struct MenuEntry *)ButtonP)->ReactNum & 0x01);
                                            break;
                                case 0x30 : BlockIndex = 0;
                                            BlockSize = GetTapeBlock (((struct MenuEntry *)ButtonP)->ReactNum & 0x01,
                                                                      &BlockIndex, &MoreBlocks);
                                            if (ViewSpectrumBASICListing (((struct MenuEntry *)ButtonP)->ReactNum & 0x01, BlockSize,
                                                         TapeWindow[((struct MenuEntry *)ButtonP)->ReactNum & 0x01].ModeIs128K))
                                              SetupNewData (((struct MenuEntry *)ButtonP)->ReactNum & 0x01);
                                            break;
                                case 0x40 : BlockIndex = 0;
                                            BlockSize = GetTapeBlock (((struct MenuEntry *)ButtonP)->ReactNum & 0x01,
                                                                      &BlockIndex, &MoreBlocks);
                                            ViewSpectrumScreenBlock (((struct MenuEntry *)ButtonP)->ReactNum & 0x01, BlockSize);
                                            break;
                                case 0x50 : RunDisassemblerOrDump ((struct MenuEntry *)ButtonP, TRUE);
                                            break;
                              }
                              break;
        case WINH_RNOHAND   : RequestQuit = TRUE; break;                           /* `Quit' is the only button without a handler */
      }
      if (RequestQuit)
      {
        if (TPConfig.ConfirmQuit)
          if (QuestionRequester (-1, -1, FALSE, FALSE, NULL, NULL, NULL, NULL, NULL, "Confirm Quit",
              "Are you sure you want to quit?"))
            RequestQuit = FALSE;
        if (RequestQuit)
          TimeToQuit = TRUE;
      }
      MouseStatus ();
    }
    CancelQuit = FALSE;
    if (TapeWindow[TWLEFT].TapeChanged)
      switch (QuestionRequester (-1, -1, TRUE, FALSE, NULL, NULL, NULL, NULL, NULL, "Confirm Quit", "Source tape has been changed! "
                                 "Write back?"))
      {
        case 0 : WriteThisTape = TRUE;                                                                                   /* (Yes) */
                 if (TapeWindow[TWLEFT].NewTape)
                   WriteThisTape = ForceNewTapeName (TWLEFT);
                 else
                   WriteThisTape = TestFileType (TWLEFT);
                 if (WriteThisTape)
                   WriteTape (TWLEFT, FALSE);
                 break;
        case 1 : break;                                                                                                   /* (No) */
        case 2 : TimeToQuit = FALSE;                                                                                  /* (Cancel) */
                 RequestQuit = FALSE;
                 CancelQuit = TRUE;
                 break;
      }
    if (TapeWindow[TWRIGHT].TapeChanged)
      switch (QuestionRequester (-1, -1, TRUE, FALSE, NULL, NULL, NULL, NULL, NULL, "Confirm Quit", "Destination tape has been "
                                 "changed! Write back?"))
      {
        case 0 : WriteThisTape = TRUE;                                                                                   /* (Yes) */
                 if (TapeWindow[TWRIGHT].NewTape)
                   WriteThisTape = ForceNewTapeName (TWRIGHT);
                 else
                   WriteThisTape = TestFileType (TWRIGHT);
                 if (WriteThisTape)
                   WriteTape (TWRIGHT, FALSE);
                 break;
        case 1 : break;                                                                                                   /* (No) */
        case 2 : TimeToQuit = FALSE;                                                                                  /* (Cancel) */
                 RequestQuit = FALSE;
                 CancelQuit = TRUE;
                 break;
      }
  }

  ReleaseTape (TWLEFT);
  ReleaseTape (TWRIGHT);
  DestroyComboWindow (&(TapeWindow[TWLEFT].IndexWindow), 0);
  DestroyComboWindow (&(TapeWindow[TWRIGHT].IndexWindow), 0);
  DestroyWindow (&(Overlay[TWLEFT]));
  DestroyWindow (&(Overlay[TWRIGHT]));
  DestroyWindow (&MainWindow);
  //ReleaseAudioBuffer ();
  EndGUI ();
  exit (0);
}
